import logging
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date
from decimal import Decimal
from typing import Final

from moneyed import GBP, Money

from investir.config import config
from investir.typing import TaxYear

logger = logging.getLogger(__name__)

TAX_YEAR_START_MONTH: Final = 4
TAX_YEAR_START_DAY: Final = 6


def date_to_tax_year(d: date) -> TaxYear:
    ty_start = date(d.year, TAX_YEAR_START_MONTH, TAX_YEAR_START_DAY)
    return TaxYear(d.year + 1) if d >= ty_start else TaxYear(d.year)


def tax_year_short_date(tax_year: TaxYear) -> str:
    return f"{tax_year - 1}/{(tax_year) % 100}"


def tax_year_full_date(tax_year: TaxYear) -> str:
    return f"6th April {tax_year - 1} to 5th April {tax_year}"


def multifilter(filters: Sequence[Callable] | None, iterable: Iterable) -> Iterable:
    if not filters:
        return iterable
    return filter(lambda x: all(f(x) for f in filters), iterable)


def raise_or_warn(ex: Exception) -> None:
    if config.strict:
        raise ex
    logger.warning(ex)


def read_decimal(val: str, default: Decimal = Decimal("0.0")) -> Decimal:
    return Decimal(val) if val.strip() else default


def read_sterling(amount: str | None) -> Money | None:
    return (
        Money(amount=amount, currency="GBP")
        if amount is not None and amount.strip()
        else None
    )


def sterling(amount: str) -> Money:
    return Money(amount=amount, currency=GBP)


def dict2str(d: Mapping[str, str]) -> str:
    return str({k: v for k, v in d.items() if v.strip()})


def boldify(text: str) -> str:
    return f"\033[1m{text}\033[0m"



import importlib.metadata
import logging
import operator
from collections.abc import Callable, Sequence
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

import click
import typer

from investir.config import config
from investir.exceptions import InvestirError
from investir.findata import (
    FinancialData,
    HistoricalExchangeRateProvider,
    HmrcMonthlyExhangeRateProvider,
    LocalHistoricalExchangeRateProvider,
    YahooFinanceHistoricalExchangeRateProvider,
    YahooFinanceLiveExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)
from investir.logging import configure_logger
from investir.output import OutputGenerator
from investir.parser import ParserFactory
from investir.prettytable import OutputFormat
from investir.taxcalculator import TaxCalculator
from investir.transaction import Acquisition, Disposal, Transaction
from investir.trhistory import TransactionHistory
from investir.typing import TaxYear, Ticker

logger = logging.getLogger(__name__)


class OrderedCommands(typer.core.TyperGroup):
    def list_commands(self, ctx: click.Context) -> list[str]:
        return list(self.commands.keys())


class MutuallyExclusiveOption(click.exceptions.UsageError):
    def __init__(self, opt1: str, opt2: str) -> None:
        super().__init__(f"Option {opt1} cannot be used together with option {opt2}")


class RatesProvider(str, Enum):
    YAHOO_FINANCE = "yahoo-finance"
    HMRC_MONTHLY = "hmrc-monthly"


app = typer.Typer(
    cls=OrderedCommands,
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    pretty_exceptions_enable=False,
)

FilesArg = Annotated[
    list[Path],
    typer.Argument(
        exists=True,
        dir_okay=False,
        help="CSV files with the account activity.",
        show_default=False,
    ),
]

TaxYearOpt = Annotated[
    Optional[int],
    typer.Option(
        min=2009,
        max=datetime.now().year + 1,
        metavar="TAX-YEAR",
        help="Filter by tax year (e.g. Use 2025 to view the transactions "
        "made in the 2024-2025 tax year).",
        show_default=False,
    ),
]

TickerOpt = Annotated[
    Optional[str],
    typer.Option(metavar="TICKER", help="Filter by ticker.", show_default=False),
]

IncludeFxFeesOpt = Annotated[
    bool, typer.Option(help="Include foreign exchange fees as an allowable cost.")
]

RatesProviderOpt = Annotated[
    RatesProvider,
    typer.Option(
        "--rates-provider",
        help="Online provider for historical exchange rates. "
        "This option is ignored if --rates-file is also used.",
    ),
]

RatesFileOpt = Annotated[
    Optional[Path],
    typer.Option(
        exists=True, dir_okay=False, help="File with historical exchange rates."
    ),
]

OutputFormatOpt = Annotated[
    OutputFormat,
    typer.Option("--output", "-o", help="Output format."),
]


def abort(msg: InvestirError | str) -> None:
    logger.critical(str(msg))
    if isinstance(msg, InvestirError) and msg.skippable:
        option = typer.style("--no-strict", fg=typer.colors.CYAN, bold=True)
        print(f"\nUse the {option} option to ignore this error.")
    raise typer.Exit(code=1)


def parse(input_files: Sequence[Path]) -> TransactionHistory:
    orders = []
    dividends = []
    transfers = []
    interest = []

    for path in input_files:
        parser, parser_name = ParserFactory.create_parser(path)
        if parser:
            logger.info("Parsing '%s' with %s parser", path, parser_name)
            try:
                result = parser.parse()
            except InvestirError as ex:
                abort(ex)

            logger.info(
                "Parsed "
                "%s orders, %s dividend payments, %s transfers, %s interest payments",
                len(result.orders),
                len(result.dividends),
                len(result.transfers),
                len(result.interest),
            )

            orders += result.orders
            dividends += result.dividends
            transfers += result.transfers
            interest += result.interest
        else:
            abort(f"Unable to find a parser for '{path}'")

    trhistory = TransactionHistory(
        orders=orders, dividends=dividends, transfers=transfers, interest=interest
    )

    if len(input_files) >= 2:
        logger.info(
            "Total: "
            "%s orders, %s dividend payments, %s transfers, %s interest payments",
            len(trhistory.orders),
            len(trhistory.dividends),
            len(trhistory.transfers),
            len(trhistory.interest),
        )

    return trhistory


def make_output_generator(
    trhistory: TransactionHistory, ctx: typer.Context
) -> OutputGenerator:
    security_info_provider = YahooFinanceSecurityInfoProvider()
    live_rates_provider = YahooFinanceLiveExchangeRateProvider()
    historical_rates_provider: HistoricalExchangeRateProvider

    if (rates_file := ctx.params.get("rates_file")) is not None:  # pragma: no cover
        historical_rates_provider = LocalHistoricalExchangeRateProvider(
            Path(rates_file)
        )
    else:
        match ctx.params.get("rates_provider"):
            case RatesProvider.YAHOO_FINANCE | None:
                historical_rates_provider = YahooFinanceHistoricalExchangeRateProvider()
            case RatesProvider.HMRC_MONTHLY:  #  pragma: no cover
                historical_rates_provider = HmrcMonthlyExhangeRateProvider()

    findata = FinancialData(
        security_info_provider, live_rates_provider, historical_rates_provider
    )
    taxcalc = TaxCalculator(trhistory, findata)

    return OutputGenerator(trhistory, taxcalc)


def create_filters(
    tax_year: int | None = None,
    ticker: str | None = None,
    tr_type: type[Transaction] | None = None,
    total_op: Callable | None = None,
) -> Sequence[Callable[[Transaction], bool]]:
    filters = []

    if tax_year is not None:
        filters.append(lambda tr: tr.tax_year() == TaxYear(tax_year))

    if ticker is not None:
        filters.append(lambda tr: tr.ticker == ticker)

    if tr_type is not None:
        filters.append(lambda tr: isinstance(tr, tr_type))

    if total_op is not None:
        filters.append(lambda tr: total_op(tr.total.amount, 0.0))

    return filters


def version_callback(value: bool) -> None:
    if value:
        print(f"{__package__} {importlib.metadata.version(__package__)}")
        raise typer.Exit()


@app.callback()
def main_callback(
    strict: Annotated[
        bool, typer.Option(help="Abort if data integrity issues are found.")
    ] = config.strict,
    offline: Annotated[
        bool,
        typer.Option(
            "--offline",
            help=(
                "Disable fetching additional data about securities (e.g. list of "
                "share sub-division events) from the Internet."
            ),
        ),
    ] = config.offline,
    cache_dir: Annotated[
        Path,
        typer.Option(
            file_okay=False,
            help=(
                "Location where to store cache files with securities information "
                "and foreign exchange rates downloaded from the Internet."
            ),
        ),
    ] = config.cache_dir,
    verbose: Annotated[
        bool, typer.Option("--verbose", help="Enable additional logging.")
    ] = False,
    quiet: Annotated[
        bool, typer.Option("--quiet", help="Disable all non-critical logging.")
    ] = False,
    colour: Annotated[
        bool, typer.Option(help="Show coloured output.")
    ] = config.use_colour,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            callback=version_callback,
            help="Show version information and exit.",
        ),
    ] = None,
) -> None:
    if verbose and quiet:
        raise MutuallyExclusiveOption("--verbose", "--quiet")

    config.strict = strict
    config.offline = offline
    config.cache_dir = cache_dir

    if quiet:
        config.log_level = logging.CRITICAL

    if verbose:
        config.log_level = logging.DEBUG

    config.use_colour = colour

    configure_logger()


@app.command("orders")
def orders_command(
    ctx: typer.Context,
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    acquisitions_only: Annotated[
        bool, typer.Option("--acquisitions", help="Show only acquisitions.")
    ] = False,
    disposals_only: Annotated[
        bool, typer.Option("--disposals", help="Show only disposals.")
    ] = False,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show share buy/sell orders.
    """
    if acquisitions_only and disposals_only:
        raise MutuallyExclusiveOption("--acquisitions", "--disposals")

    tr_type: type[Transaction] | None = None
    if acquisitions_only:
        tr_type = Acquisition
    elif disposals_only:
        tr_type = Disposal

    filters = create_filters(tax_year=tax_year, ticker=ticker, tr_type=tr_type)
    outputter = make_output_generator(parse(files), ctx)
    outputter.show_orders(format, filters)


@app.command("dividends")
def dividends_command(
    ctx: typer.Context,
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show share dividends paid out.
    """
    filters = create_filters(tax_year=tax_year, ticker=ticker)
    outputter = make_output_generator(parse(files), ctx)
    outputter.show_dividends(format, filters)


@app.command("transfers")
def transfers_command(
    ctx: typer.Context,
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    deposits_only: Annotated[
        bool, typer.Option("--deposits", help="Show only deposits.")
    ] = False,
    withdrawals_only: Annotated[
        bool, typer.Option("--withdrawals", help="Show only withdrawals.")
    ] = False,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show cash deposits and cash withdrawals.
    """
    if deposits_only and withdrawals_only:
        raise MutuallyExclusiveOption("--deposits", "--withdrawals")

    if deposits_only:
        total_op = operator.gt
    elif withdrawals_only:
        total_op = operator.lt
    else:
        total_op = None

    filters = create_filters(tax_year=tax_year, total_op=total_op)
    outputter = make_output_generator(parse(files), ctx)
    outputter.show_transfers(format, filters)


@app.command("interest")
def interest_command(
    ctx: typer.Context,
    files: FilesArg,
    tax_year: TaxYearOpt = None,
    format: OutputFormatOpt = OutputFormat.TEXT,
) -> None:
    """
    Show interest earned on cash.
    """
    filters = create_filters(tax_year=tax_year)
    outputter = make_output_generator(parse(files), ctx)
    outputter.show_interest(format, filters)


@app.command("capital-gains")
def capital_gains_command(
    ctx: typer.Context,
    files: FilesArg,
    gains_only: Annotated[
        bool, typer.Option("--gains", help="Show only capital gains.")
    ] = False,
    losses_only: Annotated[
        bool, typer.Option("--losses", help="Show only capital losses.")
    ] = False,
    tax_year: TaxYearOpt = None,
    ticker: TickerOpt = None,
    include_fx_fees: IncludeFxFeesOpt = config.include_fx_fees,
    format: OutputFormatOpt = OutputFormat.TEXT,
    rates_provider: RatesProviderOpt = RatesProvider.YAHOO_FINANCE,
    rates_file: RatesFileOpt = None,
) -> None:
    """
    Show capital gains report.
    """
    if gains_only and losses_only:
        raise MutuallyExclusiveOption("--gains", "--losses")

    if format != OutputFormat.TEXT and tax_year is None:
        raise click.exceptions.UsageError(
            f"The {format.value} format requires the option --tax-year to be used"
        )

    config.include_fx_fees = include_fx_fees

    tax_year = TaxYear(tax_year) if tax_year else None
    ticker = Ticker(ticker) if ticker else None

    try:
        outputter = make_output_generator(parse(files), ctx)
        outputter.show_capital_gains(format, tax_year, ticker, gains_only, losses_only)
    except InvestirError as ex:
        abort(ex)


@app.command("holdings")
def holdings_command(
    ctx: typer.Context,
    files: FilesArg,
    ticker: TickerOpt = None,
    show_gain_loss: Annotated[
        bool, typer.Option("--show-gain-loss", help="Show unrealised gain/loss.")
    ] = False,
    include_fx_fees: IncludeFxFeesOpt = config.include_fx_fees,
    format: OutputFormatOpt = OutputFormat.TEXT,
    rates_provider: RatesProviderOpt = RatesProvider.YAHOO_FINANCE,
    rates_file: RatesFileOpt = None,
) -> None:
    """
    Show current holdings.
    """
    config.include_fx_fees = include_fx_fees

    ticker = Ticker(ticker) if ticker else None

    try:
        outputter = make_output_generator(parse(files), ctx)
        outputter.show_holdings(format, ticker, show_gain_loss)
    except InvestirError as ex:
        abort(ex)


def main() -> None:
    app()



import logging
from collections.abc import Mapping
from csv import DictReader
from datetime import datetime, timezone
from decimal import Decimal
from pathlib import Path
from typing import Final

from dateutil.parser import parse as parse_timestamp
from moneyed import Money

from investir.const import MIN_TIMESTAMP
from investir.exceptions import (
    CalculatedAmountError,
    FeesError,
    FieldUnknownError,
    OrderDateError,
    ParseError,
    TransactionTypeError,
)
from investir.fees import Fees
from investir.parser.factory import ParserFactory
from investir.parser.types import ParsingResult
from investir.transaction import (
    Acquisition,
    Disposal,
    Dividend,
    Interest,
    Order,
    Transfer,
)
from investir.typing import ISIN, Ticker
from investir.utils import dict2str, raise_or_warn, read_decimal, read_sterling

logger = logging.getLogger(__name__)


def read_money(row: Mapping[str, str], amount_field: str) -> Money | None:
    if amount := row.get(amount_field, "").strip():
        return Money(amount=amount, currency=row[f"Currency ({amount_field})"])

    if amount := row.get(f"{amount_field} (GBP)", "").strip():
        return Money(amount=amount, currency="GBP")

    return None


@ParserFactory.register("Trading212")
class Trading212Parser:
    FIELDS: Final = (
        "Action",
        "Time",
        "Notes",
        "ID",
        "ISIN",
        "Ticker",
        "Name",
        "No. of shares",
        "Price / share",
        "Currency (Price / share)",
        "Exchange rate",
        "Total",
        "Currency (Total)",
        # Dividend
        "Withholding tax",
        "Currency (Withholding tax)",
        # Fees
        "Stamp duty (GBP)",
        "Stamp duty reserve tax (GBP)",
        "Currency conversion fee",
        "Currency (Currency conversion fee)",
        "Finra fee",
        "Currency (Finra fee)",
        "Transaction fee",
        "Currency (Transaction fee)",
        # Legacy
        "Total (GBP)",
        "Currency conversion fee (GBP)",
        "Transaction fee (GBP)",
        "Finra fee (GBP)",
        # Ignored
        "Result",
        "Currency (Result)",
        "Charge amount (GBP)",
        "Deposit fee (GBP)",
        "Currency conversion from amount",
        "Currency (Currency conversion from amount)",
        "Currency conversion to amount",
        "Currency (Currency conversion to amount)",
        "Merchant name",
        "Merchant category",
    )

    def __init__(self, csv_file: Path) -> None:
        self._csv_file = csv_file
        self._orders: list[Order] = []
        self._dividends: list[Dividend] = []
        self._transfers: list[Transfer] = []
        self._interest: list[Interest] = []

    def can_parse(self) -> bool:
        with self._csv_file.open(encoding="utf-8") as file:
            reader = DictReader(file)
            fieldnames = reader.fieldnames or []

        if "Action" not in fieldnames or "Time" not in fieldnames:
            return False

        if "Total" not in fieldnames and "Total (GBP)" not in fieldnames:
            return False

        return True

    def parse(self) -> ParsingResult:
        parse_fn = {
            "Market buy": self._parse_order,
            "Limit buy": self._parse_order,
            "Stop buy": self._parse_order,
            "Market sell": self._parse_order,
            "Limit sell": self._parse_order,
            "Stop sell": self._parse_order,
            "Dividend (Ordinary)": self._parse_dividend,
            "Dividend (Dividend)": self._parse_dividend,
            "Dividend (Dividends paid by us corporations)": self._parse_dividend,
            "Dividend (Dividends paid by foreign corporations)": self._parse_dividend,
            "Deposit": self._parse_transfer,
            "Withdrawal": self._parse_transfer,
            "Interest on cash": self._parse_interest,
            "Lending interest": self._parse_interest,
            "Result adjustment": None,
            "Card debit": None,
            "Spending cashback": None,
            "Currency conversion": None,
        }

        with self._csv_file.open(encoding="utf-8") as file:
            reader = DictReader(file)

            unknown_fields = [
                f for f in reader.fieldnames or [] if f not in self.FIELDS
            ]
            if unknown_fields:
                raise_or_warn(FieldUnknownError(unknown_fields))

            for row in reader:
                tr_type = row["Action"]

                if tr_type not in parse_fn:
                    raise_or_warn(TransactionTypeError(self._csv_file, row, tr_type))
                    continue

                if fn := parse_fn.get(tr_type):
                    timestamp = parse_timestamp(row["Time"])
                    tr_id = row["ID"]

                    if timestamp.tzinfo is None:
                        timestamp = timestamp.replace(tzinfo=timezone.utc)

                    if (total := read_money(row, "Total")) is not None:
                        fn(row, tr_type, timestamp, tr_id, total)

        return ParsingResult(
            self._orders, self._dividends, self._transfers, self._interest
        )

    def _parse_order(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        tr_id: str,
        total: Money,
    ) -> None:
        isin = row["ISIN"]
        ticker = row["Ticker"]
        name = row["Name"]
        num_shares = Decimal(row["No. of shares"])
        price_share = Decimal(row["Price / share"])
        exchange_rate = read_decimal(row["Exchange rate"], default=Decimal("1.0"))

        # TODO: The PTM levy fee and the French Financial Transaction
        #       Tax (FRR) fee are not parsed as it is unclear what their
        #       field titles would be in the CSV. The list of possible
        #       fees is available at:
        #       https://helpcentre.trading212.com/hc/en-us/articles/360007081637-What-are-the-applicable-stock-exchange-fees
        stamp_duty = read_sterling(row.get("Stamp duty (GBP)"))
        stamp_duty_reserve_tax = read_sterling(row.get("Stamp duty reserve tax (GBP)"))
        forex_fee = read_money(row, "Currency conversion fee")
        sec_fee = read_money(row, "Transaction fee")
        finra_fee = read_money(row, "Finra fee")

        if timestamp < MIN_TIMESTAMP:
            raise OrderDateError(self._csv_file, row)

        if stamp_duty and stamp_duty_reserve_tax:
            raise FeesError(
                self._csv_file, row, "Stamp duty (GBP)", "Stamp duty reserve tax (GBP)"
            )

        stamp_duty = stamp_duty or stamp_duty_reserve_tax

        if stamp_duty and finra_fee:
            raise FeesError(self._csv_file, row, "Stamp duty (GBP)", "Finra fee")

        if stamp_duty and sec_fee:
            raise FeesError(self._csv_file, row, "Stamp duty (GBP)", "Transaction fee")

        fees = Fees(
            stamp_duty=stamp_duty,
            forex=forex_fee,
            finra=finra_fee,
            sec=sec_fee,
            default_currency=total.currency,
        )

        # NB: If one of the fees is in a currency different than the currency
        # for the total or the currency for any other fee that also applies, a
        # TypeError exception will be raised below. This is not something
        # expected to happen but it is good to be prepared for that eventuality
        # instead of silently ignore it.

        order_class: type[Order] = Acquisition
        fees_total = fees.total

        if tr_type in ("Market sell", "Limit sell", "Stop sell"):
            order_class = Disposal
            fees_total *= -1

        calculated_total = (
            Money(price_share * num_shares / exchange_rate, total.currency) + fees_total
        ).round(2)

        if abs(calculated_total - total).amount > Decimal("0.01"):
            raise_or_warn(
                CalculatedAmountError(
                    self._csv_file, row, total.amount, calculated_total.amount
                )
            )

        self._orders.append(
            order_class(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=name,
                total=total,
                quantity=num_shares,
                fees=fees,
                tr_id=tr_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._orders[-1])

    def _parse_dividend(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        tr_id: str,
        total: Money,
    ):
        isin = row["ISIN"]
        ticker = row["Ticker"]
        name = row["Name"]
        currency_price_share = row["Currency (Price / share)"]
        withholding_tax = read_decimal(row["Withholding tax"])
        currency_withholding_tax = row["Currency (Withholding tax)"]
        fx_conversion_fee = row["Currency conversion fee"]

        if fx_conversion_fee:
            raise ParseError(self._csv_file, row, "Dividend with conversion fee")

        if currency_price_share != currency_withholding_tax:
            raise ParseError(
                self._csv_file,
                row,
                "Currency is different for share price and tax withheld",
            )

        self._dividends.append(
            Dividend(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=name,
                total=total,
                withheld=Money(withholding_tax, currency_withholding_tax),
                tr_id=tr_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._dividends[-1])

    def _parse_transfer(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        tr_id: str,
        total: Money,
    ):
        if tr_type == "Withdrawal":
            total = -abs(total)

        self._transfers.append(Transfer(timestamp, tr_id=tr_id, total=total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._transfers[-1])

    def _parse_interest(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        tr_id: str,
        total: Money,
    ):
        self._interest.append(Interest(timestamp, tr_id=tr_id, total=total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._interest[-1])



import logging
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from decimal import Decimal
from typing import TypeVar

from investir.config import config
from investir.const import BASE_CURRENCY
from investir.exceptions import AmbiguousTickerError
from investir.prettytable import Field, Format, OutputFormat, PrettyTable
from investir.taxcalculator import TaxCalculator
from investir.transaction import (
    Acquisition,
    Disposal,
    Transaction,
)
from investir.trhistory import TransactionHistory
from investir.typing import TaxYear, Ticker
from investir.utils import boldify, multifilter, tax_year_full_date, tax_year_short_date

T = TypeVar("T", bound=Transaction)

logger = logging.getLogger(__name__)


def gbp(amount: Decimal) -> str:
    sign = "" if amount >= 0.0 else "-"
    return f"{sign}£{abs(amount):.2f}"


@dataclass
class CapitalGainsSummary:
    num_disposals: int
    disposal_proceeds: Decimal
    total_cost: Decimal
    total_gains: Decimal
    total_losses: Decimal

    @property
    def net_gains(self) -> Decimal:
        return self.total_gains - self.total_losses

    def __str__(self) -> str:
        return (
            f"{'Number of disposals:':40}{self.num_disposals:>10}      "
            f"{'Gains in the year, before losses:':34}{gbp(self.total_gains):>10}\n"
            f"{'Disposal proceeds:':40}{gbp(self.disposal_proceeds):>10}      "
            f"{'Losses in the year:':34}{gbp(self.total_losses):>10}\n"
            f"{'Allowable costs (incl. purchase price):':40}"
            f"{gbp(self.total_cost):>10}      "
            f"{'Net gain or loss:':34}{gbp(self.net_gains):>10}\n"
        )


class OutputGenerator:
    def __init__(self, trhistory: TransactionHistory, taxcalc: TaxCalculator) -> None:
        self._trhistory = trhistory
        self._taxcalc = taxcalc

    def show_orders(
        self, format: OutputFormat, filters: Sequence[Callable] | None = None
    ) -> None:
        if table := self._create_orders_table(filters):
            print(table.to_string(format, leading_nl=config.logging_enabled))

    def show_dividends(
        self, format: OutputFormat, filters: Sequence[Callable] | None = None
    ) -> None:
        if table := self._create_dividends_table(filters):
            print(table.to_string(format, leading_nl=config.logging_enabled))

    def show_transfers(
        self, format: OutputFormat, filters: Sequence[Callable] | None = None
    ) -> None:
        if table := self._create_transfers_table(filters):
            print(table.to_string(format, leading_nl=config.logging_enabled))

    def show_interest(
        self, format: OutputFormat, filters: Sequence[Callable] | None = None
    ) -> None:
        if table := self._create_interest_table(filters):
            print(table.to_string(format, leading_nl=config.logging_enabled))

    def show_capital_gains(
        self,
        format: OutputFormat,
        tax_year_filter: TaxYear | None,
        ticker_filter: Ticker | None,
        gains_only: bool,
        losses_only: bool,
    ) -> None:
        if tax_year_filter is not None:
            tax_years: Sequence = [tax_year_filter]
        else:
            tax_years = sorted(self._taxcalc.disposal_years())

        for tax_year_idx, tax_year in enumerate(tax_years):
            table, summary = self._create_capital_gains_table_and_summary(
                tax_year, ticker_filter, gains_only, losses_only
            )

            if table:
                print(end="\n" if tax_year_idx == 0 and config.logging_enabled else "")

                if format == OutputFormat.TEXT:
                    print(
                        boldify(
                            f"Capital Gains Tax Report {tax_year_short_date(tax_year)}"
                        )
                    )
                    print(tax_year_full_date(tax_year))
                    print(table.to_string(format))
                    print(summary)
                else:
                    print(table.to_string(format, leading_nl=False))

    def show_holdings(
        self,
        format: OutputFormat,
        ticker_filter: Ticker | None = None,
        show_gain_loss: bool = False,
    ) -> None:
        if table := self._create_holdings_table(ticker_filter, show_gain_loss):
            print(table.to_string(format, leading_nl=config.logging_enabled))

    def _create_orders_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            [
                Field("Date", Format.DATE),
                Field("Security Name"),
                Field("ISIN"),
                Field("Ticker"),
                Field("Cost", Format.MONEY, show_sum=True),
                Field("Proceeds", Format.MONEY, show_sum=True),
                Field("Quantity", Format.QUANTITY),
                Field("Price", Format.MONEY),
                Field("Fees", Format.MONEY, show_sum=True),
            ]
        )

        transactions = list(multifilter(filters, self._trhistory.orders))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            cost = tr.total if isinstance(tr, Acquisition) else None
            proceeds = tr.total if isinstance(tr, Disposal) else None

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [
                    tr.date,
                    tr.name,
                    tr.isin,
                    tr.ticker,
                    cost,
                    proceeds,
                    tr.quantity,
                    tr.price,
                    tr.fees.total,
                ],
                divider=divider,
            )

        return table

    def _create_dividends_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            [
                Field("Date", Format.DATE),
                Field("Security Name"),
                Field("ISIN"),
                Field("Ticker"),
                Field("Net Amount", Format.MONEY, show_sum=True),
                Field("Widthheld Amount", Format.MONEY, show_sum=True),
            ]
        )

        transactions = list(multifilter(filters, self._trhistory.dividends))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row(
                [
                    tr.date,
                    tr.name,
                    tr.isin,
                    tr.ticker,
                    tr.total,
                    tr.withheld,
                ],
                divider=divider,
            )

        return table

    def _create_transfers_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            [
                Field("Date", Format.DATE),
                Field("Deposit", Format.MONEY, show_sum=True),
                Field("Withdrawal", Format.MONEY, show_sum=True),
            ]
        )

        transactions = list(multifilter(filters, self._trhistory.transfers))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            if tr.total.amount > 0:
                deposited = tr.total
                widthdrew = ""
            else:
                deposited = ""
                widthdrew = abs(tr.total)

            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, deposited, widthdrew], divider=divider)

        return table

    def _create_interest_table(
        self, filters: Sequence[Callable] | None = None
    ) -> PrettyTable:
        table = PrettyTable(
            [
                Field("Date", Format.DATE),
                Field("Amount", Format.MONEY, show_sum=True),
            ]
        )

        transactions = list(multifilter(filters, self._trhistory.interest))
        last_idx = len(transactions) - 1

        for idx, tr in enumerate(transactions):
            divider = (
                idx == last_idx or tr.tax_year() != transactions[idx + 1].tax_year()
            )

            table.add_row([tr.date, tr.total], divider=divider)

        return table

    def _create_capital_gains_table_and_summary(
        self,
        tax_year: TaxYear,
        ticker_filter: Ticker | None,
        gains_only: bool,
        losses_only: bool,
    ) -> tuple[PrettyTable, CapitalGainsSummary]:
        assert not (gains_only and losses_only)

        table = PrettyTable(
            [
                Field("Disposal Date", Format.DATE),
                Field("Identification"),
                Field("Security Name"),
                Field("ISIN"),
                Field("Quantity", Format.QUANTITY),
                Field(f"Cost ({BASE_CURRENCY})", Format.DECIMAL),
                Field(f"Proceeds ({BASE_CURRENCY})", Format.DECIMAL),
                Field(f"Gain/loss ({BASE_CURRENCY})", Format.DECIMAL),
            ]
        )

        num_disposals = 0
        disposal_proceeds = total_cost = total_gains = total_losses = Decimal("0.0")

        for cg in self._taxcalc.capital_gains(tax_year):
            if ticker_filter is not None and cg.disposal.ticker != ticker_filter:
                continue

            if gains_only and cg.gain_loss < 0.0:
                continue

            if losses_only and cg.gain_loss > 0.0:
                continue

            table.add_row(
                [
                    cg.disposal.date,
                    cg.identification,
                    cg.disposal.name,
                    cg.disposal.isin,
                    cg.quantity,
                    cg.cost,
                    cg.disposal.gross_proceeds.amount,
                    cg.gain_loss,
                ]
            )

            num_disposals += 1
            disposal_proceeds += round(cg.disposal.gross_proceeds.amount, 2)
            total_cost += round(cg.cost, 2)
            if cg.gain_loss > 0.0:
                total_gains += cg.gain_loss
            else:
                total_losses += abs(cg.gain_loss)

        summary = CapitalGainsSummary(
            num_disposals, disposal_proceeds, total_cost, total_gains, total_losses
        )

        return table, summary

    def _create_holdings_table(
        self, ticker_filter: Ticker | None = None, show_gain_loss: bool = False
    ) -> PrettyTable:
        table = PrettyTable(
            [
                Field("Security Name"),
                Field("ISIN"),
                Field(f"Cost ({BASE_CURRENCY})", Format.DECIMAL),
                Field("Quantity", Format.QUANTITY),
                Field(
                    f"Current Value ({BASE_CURRENCY})",
                    Format.DECIMAL,
                    visible=show_gain_loss,
                ),
                Field(
                    f"Gain/Loss ({BASE_CURRENCY})",
                    Format.DECIMAL,
                    visible=show_gain_loss,
                ),
                Field("Weight (%)", Format.DECIMAL, visible=show_gain_loss),
            ]
        )

        holdings = self._taxcalc.holdings

        if ticker_filter is not None:
            try:
                isin = self._trhistory.get_ticker_isin(ticker_filter)
            except AmbiguousTickerError as e:
                logger.warning(e)
                holdings = []
            else:
                if holding := next((h for h in holdings if h.isin == isin), None):
                    holdings = [holding]
                else:
                    holdings = []

        holdings = sorted(holdings, key=lambda h: h.cost, reverse=True)

        if show_gain_loss:
            values = {
                holding.isin: value
                for holding in holdings
                if (value := self._taxcalc.get_holding_value(holding.isin)) is not None
            }
        else:
            values = {}

        portfolio_value = sum(val for val in values.values())
        last_idx = len(holdings) - 1

        for idx, holding in enumerate(holdings):
            gain_loss: Decimal | None = None
            weight: Decimal | None = None

            if value := values.get(holding.isin):
                gain_loss = value - holding.cost
                weight = value / portfolio_value * 100

            table.add_row(
                [
                    self._trhistory.get_security_name(holding.isin),
                    holding.isin,
                    holding.cost,
                    holding.quantity,
                    value,
                    gain_loss or "n/a",
                    weight or "n/a",
                ],
                divider=idx == last_idx,
            )

        return table



import logging
from collections import defaultdict, namedtuple
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, replace
from datetime import date, timedelta
from decimal import Decimal
from typing import Any, TypeAlias, TypeVar, cast

from moneyed import GBP, Money

from investir.config import config
from investir.const import BASE_CURRENCY
from investir.exceptions import IncompleteRecordsError, InvestirError
from investir.fees import scalar_op
from investir.findata import FinancialData
from investir.transaction import Acquisition, Disposal, Order
from investir.trhistory import TransactionHistory
from investir.typing import ISIN, TaxYear
from investir.utils import raise_or_warn

logger = logging.getLogger(__name__)

FuncType = TypeVar("FuncType", bound=Callable[..., Any])

GroupKey = namedtuple("GroupKey", ["isin", "date", "type"])
GroupDict: TypeAlias = Mapping[GroupKey, Sequence[Order]]


def same_day_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord1.date == ord2.date


def thirty_days_match(ord1: Acquisition, ord2: Disposal) -> bool:
    assert ord1.isin == ord2.isin
    return ord2.date < ord1.date <= ord2.date + timedelta(days=30)


@dataclass
class CapitalGain:
    disposal: Disposal
    cost: Decimal
    acquisition_date: date | None = None

    @property
    def gain_loss(self) -> Decimal:
        # Disposal fees have been added to the `cost` so the gross
        # proceeds are used.
        return (
            self.disposal.gross_proceeds
            - Money(self.cost, self.disposal.total.currency)
        ).amount

    @property
    def quantity(self) -> Decimal:
        if self.disposal.original_quantity is not None:
            return self.disposal.original_quantity
        return self.disposal.quantity

    @property
    def identification(self) -> str:
        match self.acquisition_date:
            case None:
                return "Section 104"
            case self.disposal.date:
                return "Same day"
            case _:
                return f"Bed & B. ({self.acquisition_date})"

    def __str__(self) -> str:
        return (
            f"{self.disposal.date} "
            f"{self.disposal.isin:<4} "
            f"quantity: {self.quantity}, "
            f"cost: £{self.cost:.2f}, "
            f"proceeds: £{self.disposal.gross_proceeds.amount}, "
            f"gain: £{self.gain_loss:.2f}, "
            f"identification: {self.identification}"
        )


@dataclass
class Section104Holding:
    isin: ISIN
    cost: Decimal
    quantity: Decimal

    def add(self, cost: Decimal, quantity: Decimal) -> None:
        self.cost += cost
        self.quantity += quantity

    def remove(self, cost: Decimal, quantity: Decimal) -> None:
        self.cost -= cost
        self.quantity -= quantity


def lazycalc(func: FuncType) -> FuncType:
    def _wrapper(self, *args, **kwargs):
        if not self._capital_gains and not self._holdings:
            self._calculate_capital_gains()
        return func(self, *args, **kwargs)

    return cast(FuncType, _wrapper)


class TaxCalculator:
    def __init__(self, trhistory: TransactionHistory, findata: FinancialData) -> None:
        self._trhistory = trhistory
        self._findata = findata
        self._acquisitions: dict[ISIN, list[Acquisition]] = defaultdict(list)
        self._disposals: dict[ISIN, list[Disposal]] = defaultdict(list)
        self._holdings: dict[ISIN, Section104Holding] = {}
        self._capital_gains: dict[TaxYear, list[CapitalGain]] = defaultdict(list)

    @lazycalc
    def capital_gains(self, tax_year: TaxYear | None = None) -> Sequence[CapitalGain]:
        if tax_year is not None:
            return self._capital_gains.get(tax_year, [])

        return [cg for cg_group in self._capital_gains.values() for cg in cg_group]

    @property
    @lazycalc
    def holdings(self) -> Sequence[Section104Holding]:
        return list(self._holdings.values())

    @lazycalc
    def holding(self, isin: ISIN) -> Section104Holding | None:
        return self._holdings.get(isin)

    @lazycalc
    def get_holding_value(self, isin: ISIN) -> Decimal | None:
        holding = self._holdings[isin]
        security_name = self._trhistory.get_security_name(isin) or ""
        if (price := self._findata.get_security_price(isin, security_name)) and (
            price_base_currency := self._findata.convert_money(price, BASE_CURRENCY)
        ):
            return holding.quantity * price_base_currency.amount

        return None

    @lazycalc
    def disposal_years(self) -> Sequence[TaxYear]:
        return list(self._capital_gains.keys())

    def _calculate_capital_gains(self) -> None:
        logger.info("Calculating capital gains")

        orders = self._trhistory.orders

        # Convert acquisitions and disposals not realised in pound
        # sterling to use monetary values specified in this currency.
        orders = [self._convert_to_sterling(o) for o in orders]

        # Exclude forex fees from the `total` and `fees` fields if
        # necessary.
        if config.include_fx_fees is False:
            orders = [self._exclude_fx_fees(o) for o in orders]

        # Normalise the orders by retroactively adjusting their share
        # quantity for any eventual share sub-division or share
        # consolidation event.
        orders = [self._normalise_quantity(o) for o in orders]

        # Group together orders that have the same isin, date and type.
        same_day = self._group_same_day(orders)

        for isin, name in self._trhistory.securities:
            logger.debug("Calculating capital gains for %s (%s)", name, isin)

            # Merge orders that were issued in the same day and have
            # the same type, and place them in the acquisitions or
            # disposals bucket.
            self._merge_same_day(isin, same_day)

            # Match disposed shares with shares acquired in the same
            # day.
            self._match_shares(isin, same_day_match)

            # Match disposed shares with shares acquired up to 30 days
            # past the disposal date.
            self._match_shares(isin, thirty_days_match)

            # Process shares disposed from a Section 104 pool.
            self._process_section104_disposals(isin)

        # Capital gains are calculated ticker by ticker in order to be
        # able to show all the intermediary calculations grouped together
        # as a future improvement. However, for the report, it is more
        # intuitive to show the capital gain transactions ordered by
        # their disposal date.
        for year, events in self._capital_gains.items():
            self._capital_gains[year] = sorted(
                events, key=lambda te: (te.disposal.timestamp, te.disposal.isin)
            )

    def _convert_to_sterling(self, order: Order) -> Order:
        def convert(m):
            if m is None:
                return m
            if (new_m := self._findata.convert_money(m, GBP, order.date)) is None:
                raise InvestirError(
                    f"Failed to get historical exchange rate for "
                    f"GBP-{m.currency} on {order.date}"
                )
            return new_m

        try:
            fees_currency = order.fees.total.currency
        except TypeError:
            # Multiple fees in different currencies: conversion needed.
            fees_currency = None

        if (order.total.currency, fees_currency) != (GBP, GBP):
            total = convert(order.total)
            fees = scalar_op(convert, order.fees)
            return replace(
                order,
                total=total,
                fees=fees,
                notes="Conversion to sterling from order {order.number}",
            )

        return order

    def _exclude_fx_fees(self, order: Order) -> Order:
        if order.fees.forex:
            if isinstance(order, Acquisition):
                total = order.total - order.fees.forex
            else:
                total = order.total + order.fees.forex

            return replace(
                order,
                total=total,
                fees=replace(order.fees, forex=None),
                notes="Forex fees removed from order {order.number}",
            )

        return order

    def _normalise_quantity(self, order: Order) -> Order:
        security_info = self._findata.get_security_info(
            order.isin, order.name, order.timestamp
        )
        return order.adjust_quantity(security_info.splits)

    def _group_same_day(self, orders: Sequence[Order]) -> GroupDict:
        same_day = defaultdict(list)
        for o in orders:
            key = GroupKey(o.isin, o.date, type(o))
            same_day[key].append(o)
        return same_day

    def _merge_same_day(self, isin: ISIN, same_day: GroupDict) -> None:
        security_orders = (
            orders for key, orders in same_day.items() if key.isin == isin
        )

        for orders in security_orders:
            if len(orders) > 1:
                order = Order.merge(*orders)
                logger.debug('    New "same-day" merged order: %s', order)
            else:
                order = orders[0]

            if isinstance(order, Acquisition):
                self._acquisitions[isin].append(order)
            elif isinstance(order, Disposal):
                self._disposals[isin].append(order)

    def _match_shares(
        self, isin: ISIN, match_fn: Callable[[Acquisition, Disposal], bool]
    ):
        acquisits = self._acquisitions[isin]
        disposals = self._disposals[isin]
        matched: set[Order] = set()

        a_idx = 0
        d_idx = 0

        while d_idx < len(disposals):
            if a_idx == len(acquisits):
                a_idx = 0
                d_idx += 1
                continue

            a = acquisits[a_idx]
            d = disposals[d_idx]

            if not match_fn(a, d) or a in matched:
                a_idx += 1
                continue

            matched.add(a)
            matched.add(d)

            if a.quantity > d.quantity:
                a, acquisits[a_idx] = a.split(d.quantity)
                a_idx = 0
                d_idx += 1
            elif d.quantity > a.quantity:
                d, disposals[d_idx] = d.split(a.quantity)
                a_idx += 1
            else:
                a_idx = 0
                d_idx += 1

            self._capital_gains[d.tax_year()].append(
                CapitalGain(d, a.total.amount + d.fees.total.amount, a.date)
            )

        self._acquisitions[isin] = [o for o in acquisits if o not in matched]
        self._disposals[isin] = [o for o in disposals if o not in matched]

    def _process_section104_disposals(self, isin: ISIN) -> None:
        security_orders = sorted(
            [*self._acquisitions[isin], *self._disposals[isin]],
            key=lambda order: order.date,
        )

        for order in security_orders:
            holding = self._holdings.get(isin)

            if isinstance(order, Acquisition):
                if holding is not None:
                    holding.add(order.total.amount, order.quantity)
                else:
                    self._holdings[isin] = Section104Holding(
                        isin, order.total.amount, order.quantity
                    )
            elif isinstance(order, Disposal):
                if holding is not None:
                    allowable_cost = holding.cost * order.quantity / holding.quantity

                    holding.remove(allowable_cost, order.quantity)

                    if holding.quantity < 0.0:
                        raise_or_warn(
                            IncompleteRecordsError(
                                isin, self._trhistory.get_security_name(isin) or "?"
                            )
                        )
                        logger.warning("Not calculating holding for %s", isin)
                        del self._holdings[isin]
                        break

                    if holding.quantity == Decimal("0.0"):
                        del self._holdings[isin]

                    self._capital_gains[order.tax_year()].append(
                        CapitalGain(order, allowable_cost + order.fees.total.amount)
                    )
                else:
                    raise_or_warn(
                        IncompleteRecordsError(
                            isin, self._trhistory.get_security_name(isin) or "?"
                        )
                    )
                    logger.warning("Not calculating holding for %s", isin)
                    break



from datetime import datetime, timezone
from decimal import Decimal

from investir.findata import FinancialData
from investir.output import OutputGenerator
from investir.prettytable import OutputFormat
from investir.taxcalculator import TaxCalculator
from investir.transaction import (
    Acquisition,
)
from investir.trhistory import TransactionHistory
from investir.typing import ISIN, Ticker
from investir.utils import sterling

# NB: Most of the output generated by the output module is already
# tested on test_cli.py.


def test_create_holdings_table_ambiguous_ticker(capsys):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        ticker="TICKER",
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    trhistory = TransactionHistory(orders=[order1, order2])
    findata = FinancialData(None, None, None)
    taxcalc = TaxCalculator(trhistory, findata)

    outputter = OutputGenerator(trhistory, taxcalc)
    outputter.show_holdings(format=OutputFormat.TEXT, ticker_filter=Ticker("TICKER"))
    captured = capsys.readouterr()

    assert not captured.out
    assert not captured.err



from collections.abc import Callable, Sequence
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal

import pytest
from moneyed import GBP, USD, Money

from investir.config import config
from investir.exceptions import IncompleteRecordsError, InvestirError
from investir.findata import (
    FinancialData,
    RequestError,
    SecurityInfo,
    Split,
    YahooFinanceHistoricalExchangeRateProvider,
    YahooFinanceLiveExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)
from investir.taxcalculator import TaxCalculator
from investir.transaction import Acquisition, Disposal, Fees, Order
from investir.trhistory import TransactionHistory
from investir.typing import ISIN, Ticker
from investir.utils import sterling


@pytest.fixture
def make_tax_calculator(mocker, tmp_path) -> Callable:
    def _wrapper(
        orders: Sequence[Order],
        splits: Sequence[Split] | None = None,
        price: Money | Exception | None = None,
        live_rate: Decimal | Exception | None = None,
        historical_rates: Sequence[Decimal] | Exception | None = None,
    ) -> TaxCalculator:
        if splits is None:
            splits = []
        else:
            assert all(orders[0].isin == o.isin for o in orders)

        security_info_provider = YahooFinanceSecurityInfoProvider()
        live_rates_provider = YahooFinanceLiveExchangeRateProvider()
        historical_rates_provider = YahooFinanceHistoricalExchangeRateProvider()

        mocker.patch.object(
            security_info_provider,
            "get_info",
            return_value=SecurityInfo(splits=splits),
        )
        mocker.patch.object(
            security_info_provider,
            "get_price",
            side_effect=[price],
        )
        mocker.patch.object(live_rates_provider, "get_rate", side_effect=[live_rate])
        mocker.patch.object(
            historical_rates_provider, "get_rate", side_effect=historical_rates
        )

        trhistory = TransactionHistory(orders=orders)
        findata = FinancialData(
            security_info_provider, live_rates_provider, historical_rates_provider
        )

        return TaxCalculator(trhistory, findata)

    return _wrapper


def test_section_104_disposal(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1400"),
        quantity=Decimal("50.0"),
    )

    order2 = Acquisition(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1600"),
        quantity=Decimal("50.0"),
    )

    order3 = Disposal(
        datetime(2023, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("4000.0"),
        quantity=Decimal("20.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1
    assert capital_gains == taxcalc.capital_gains(2024)

    cg = capital_gains[0]
    assert cg.disposal.date == date(2023, 5, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("600.0")
    assert cg.gain_loss == Decimal("3400.0")
    assert str(cg) == (
        "2023-05-01 X    quantity: 20.0, cost: £600.00"
        ", proceeds: £4000.0, gain: £3400.00"
        ", identification: Section 104"
    )

    holding = taxcalc.holding(Ticker("X"))
    assert holding.quantity == Decimal("80.0")
    assert holding.cost == Decimal("2400.0")


def test_section_104_with_no_disposal_made(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1400.0"),
        quantity=Decimal("50.0"),
    )

    order2 = Acquisition(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1600.0"),
        quantity=Decimal("50.0"),
    )

    taxcalc = make_tax_calculator([order1, order2])
    assert not taxcalc.capital_gains()
    holding = taxcalc.holding(Ticker("X"))
    assert holding.quantity == Decimal("100.0")
    assert holding.cost == Decimal("3000.0")


def test_same_day_rule(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, 14, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("70.0"),
        quantity=Decimal("1.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 20, 15, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("60.0"),
        quantity=Decimal("1.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 20, 16, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("65.0"),
        quantity=Decimal("2.0"),
    )

    order5 = Disposal(
        datetime(2019, 1, 20, 17, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("280.0"),
        quantity=Decimal("4.0"),
    )

    order6 = Acquisition(
        datetime(2019, 1, 20, 18, 0, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("55.0"),
        quantity=Decimal("2.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5, order6])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.disposal.total == order2.total + order5.total
    assert cg.cost == order3.total.amount + order4.total.amount + order6.total.amount
    assert cg.gain_loss == Decimal("170.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("100.0")
    assert holding.quantity == Decimal("10.0")


@pytest.mark.parametrize("days_elapsed", range(1, 31))
def test_bed_and_breakfast_rule(days_elapsed, make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Acquisition(
        order2.timestamp + timedelta(days=days_elapsed),
        isin=ISIN("X"),
        total=sterling("120.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == order2.date + timedelta(days=days_elapsed)
    assert cg.cost == Decimal("120.0")
    assert cg.gain_loss == Decimal("30.0")
    assert str(cg) == (
        f"2019-01-20 X    quantity: 5.0, cost: £120.00"
        f", proceeds: £150.0, gain: £30.00"
        f", identification: Bed & B. ({cg.acquisition_date})"
    )

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("100.0")
    assert holding.quantity == Decimal("10.0")


def test_acquisitions_are_not_matched_after_thirty_days_of_disposal_date(
    make_tax_calculator,
):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Acquisition(
        datetime(2019, 2, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("120.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("50.0")
    assert cg.gain_loss == Decimal("100.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("170.0")
    assert holding.quantity == Decimal("10.0")


def test_acquisitions_are_not_matched_before_disposal_date(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Acquisition(
        datetime(2019, 2, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("200.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Disposal(
        datetime(2019, 2, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("100.0")
    assert cg.gain_loss == Decimal("50.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("200.0")
    assert holding.quantity == Decimal("10.0")


def test_same_day_rule_has_priority_to_bed_and_breakfast_rule(make_tax_calculator):
    """
    Verify that the same day rule has priority over the bed and
    breakfast rule.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    order4 = Disposal(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("170.0"),
        quantity=Decimal("5.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("300.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]  # Bed and breakfast disposal event
    assert cg.acquisition_date == date(2019, 1, 27)
    assert cg.cost == Decimal("300.0")
    assert cg.gain_loss == Decimal("-150.0")

    cg = capital_gains[1]  # Same day disposal event
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("150.0")
    assert cg.gain_loss == Decimal("20.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("100.0")
    assert holding.quantity == Decimal("10.0")


def test_matching_disposals_with_larger_acquisition(make_tax_calculator):
    """
    Test splitting an acquisition in two to match the first disposal,
    and then using the remaining part to match the second disposal.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("60.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("11.0"),
        quantity=Decimal("1.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("70.0"),
        quantity=Decimal("7.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("50.0")
    assert cg.gain_loss == Decimal("10.0")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("10.0")
    assert cg.gain_loss == Decimal("1.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.quantity == Decimal("11.0")
    assert holding.cost == Decimal("110.0")


def test_matching_disposal_with_multiple_smaller_acquisitions(make_tax_calculator):
    """
    Test multiple tax events derived from a single disposal
       - 1 share matched due same day rule
       - 2 shares matched due bed and breakfast rule on 2019/1/25
       - 1 shares matched due bed and breakfast rule on 2019/1/27
       - 1 share from the Section 104 pool

    In addition, verify that the second disposal does not match
    against acquisitions that were already matched for the first
    disposal.
    """
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("30.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("50.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("9.0"),
        quantity=Decimal("1.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("16.0"),
        quantity=Decimal("2.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("5.0"),
        quantity=Decimal("1.0"),
    )

    order6 = Disposal(
        datetime(2019, 3, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("7.0"),
        quantity=Decimal("1.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5, order6])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 5

    cg = capital_gains[0]  # Same day
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.cost == Decimal("9.0")
    assert cg.gain_loss == Decimal("1.0")

    cg = capital_gains[1]  # First bed and breakfast match
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("16.0")
    assert cg.gain_loss == Decimal("4.0")

    cg = capital_gains[2]  # Second bed and breakfast match
    assert cg.acquisition_date == date(2019, 1, 27)
    assert cg.cost == Decimal("5.0")
    assert cg.gain_loss == Decimal("5.0")

    cg = capital_gains[3]  # Section 104 pool
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("3.0")
    assert cg.gain_loss == Decimal("7.0")

    cg = capital_gains[4]  # Section 104 pool
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("3.0")
    assert cg.gain_loss == Decimal("4.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("24.0")
    assert holding.quantity == Decimal("8.0")


def test_capital_gains_on_orders_with_fees_incurred(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("31.5"),
        quantity=Decimal("10.0"),
        fees=Fees(stamp_duty=sterling("1.5")),
    )

    order2 = Acquisition(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("40.5"),
        quantity=Decimal("5.0"),
        fees=Fees(stamp_duty=sterling("0.5")),
    )

    order3 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("49.6"),
        quantity=Decimal("5.0"),
        fees=Fees(stamp_duty=sterling("0.4")),
    )

    order4 = Disposal(
        datetime(2019, 3, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("49.2"),
        quantity=Decimal("5.0"),
        fees=Fees(stamp_duty=sterling("0.8")),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    # Acquisition cost = 40.0 + 0.4 + 0.5 = 40.9
    # Gain/Loss        = 50.0 - 40.9 = 33.45 = 9.1
    cg = capital_gains[1]
    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 20)
    assert cg.cost == Decimal("40.9")
    assert cg.gain_loss == Decimal("9.1")

    # S104 pool cost   = 30.0 + 1.5 = 31.5
    # Acquisition cost = 31.5 * (5.0/10.0) + 0.8 = 16.55
    # Gain/Loss        = 50.0 - 16.55 = 33.45
    cg = capital_gains[1]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("16.55")
    assert cg.gain_loss == Decimal("33.45")

    # New S104 pool cost = 31.5 - 31.5 * (5.0/10.0) = 15.75
    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("15.75")
    assert holding.quantity == Decimal("5.0")


def test_disposals_on_different_tickers(make_tax_calculator):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        total=sterling("200.0"),
        quantity=Decimal("20.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("11.0"),
        quantity=Decimal("1.0"),
    )

    order4 = Disposal(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        total=sterling("22.0"),
        quantity=Decimal("2.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 21, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("8.0"),
        quantity=Decimal("2.0"),
    )

    order6 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("Y"),
        total=sterling("6.0"),
        quantity=Decimal("2.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5, order6])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 21)
    assert cg.disposal.isin == ISIN("X")
    assert cg.cost == Decimal("4.0")
    assert cg.gain_loss == Decimal("7.0")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.disposal.isin == ISIN("Y")
    assert cg.cost == Decimal("6.0")
    assert cg.gain_loss == Decimal("16.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("104.0")
    assert holding.quantity == Decimal("11.0")

    holding = taxcalc.holding(Ticker("Y"))
    assert holding.cost == Decimal("200.0")
    assert holding.quantity == Decimal("20.0")


def test_integrity_disposal_without_acquisition(make_tax_calculator):
    order = Disposal(
        datetime(2019, 1, 17, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("120.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator([order])
    with pytest.raises(IncompleteRecordsError):
        taxcalc.capital_gains()

    taxcalc = make_tax_calculator([order])
    config.strict = False
    taxcalc.capital_gains()


def test_integrity_disposing_more_than_acquired(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("5.0"),
    )

    order2 = Disposal(
        datetime(2019, 3, 17, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("120.0"),
        quantity=Decimal("10.0"),
    )

    taxcalc = make_tax_calculator([order1, order2])
    with pytest.raises(IncompleteRecordsError):
        taxcalc.capital_gains()

    taxcalc = make_tax_calculator([order1, order2])
    config.strict = False
    taxcalc.capital_gains()


def test_capital_gains_on_orders_not_realised_in_pound_sterling(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=Money("1000.0", "USD"),
        quantity=Decimal("10.0"),
        fees=Fees(finra=Money("0.5", "USD")),
    )

    order2 = Disposal(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=Money("1500.0", "USD"),
        quantity=Decimal("5.0"),
        fees=Fees(finra=Money("0.6", "USD")),
    )

    taxcalc = make_tax_calculator(
        [order1, order2],
        historical_rates=[
            Decimal("0.775"),
            Decimal("0.775"),
            Decimal("0.76"),
            Decimal("0.76"),
        ],
    )

    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.cost == Decimal("387.9560")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("752.5000")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("387.5")
    assert holding.quantity == Decimal("5.0")

    taxcalc = make_tax_calculator(
        [order1, order2],
        historical_rates=RequestError,
    )

    with pytest.raises(InvestirError):
        taxcalc.capital_gains()


def test_capital_gains_on_orders_not_realised_in_pound_sterling_when_rate_not_available(
    make_tax_calculator,
):
    order = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=Money("1000.0", "USD"),
        quantity=Decimal("10.0"),
    )

    taxcalc = make_tax_calculator(
        [order],
        historical_rates=RequestError,
    )

    with pytest.raises(InvestirError):
        taxcalc.capital_gains()


def test_capital_gains_on_orders_with_fees_in_different_currencies(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1000.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1500.0"),
        quantity=Decimal("5.0"),
        fees=Fees(finra=Money("1.5", "USD"), sec=Money("2.4", "EUR")),
    )

    taxcalc = make_tax_calculator(
        [order1, order2],
        historical_rates=[
            Decimal("0.775"),
            Decimal("0.854"),
        ],
    )

    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    # Disposal fees are added to the cost. So
    # Cost (GBP) = 500 + (1.5 * 0.775) + (2.4 * 0.854)
    assert cg.cost == Decimal("503.2121")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("1000.0")


def test_capital_gains_on_orders_with_forex_fees_excluded(
    make_tax_calculator,
):
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("52.0"),
        quantity=Decimal("5.0"),
        fees=Fees(forex=sterling("2.0")),
    )

    order2 = Acquisition(
        datetime(2018, 1, 2, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("50.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("197.0"),
        quantity=Decimal("5.0"),
        fees=Fees(forex=sterling("3.0")),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    config.include_fx_fees = False
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("50.0")
    assert cg.gain_loss == Decimal("150.0")


def test_section_104_disposal_with_share_split(make_tax_calculator):
    order1 = Acquisition(
        datetime(2014, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("3300.0"),
        quantity=Decimal("11.0"),
    )

    order2 = Disposal(
        datetime(2014, 6, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("500.0"),
        quantity=Decimal("1.0"),
    )

    order3 = Acquisition(
        datetime(2014, 8, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1000.0"),
        quantity=Decimal("10.0"),
    )

    order4 = Disposal(
        datetime(2014, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("1100.0"),
        quantity=Decimal("5.0"),
    )

    taxcalc = make_tax_calculator(
        [order1, order2, order3, order4],
        [Split(datetime(2014, 7, 1, tzinfo=timezone.utc), Decimal("3.0"))],
    )
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2014, 6, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("300.0")
    assert cg.quantity == Decimal("1.0")
    assert cg.gain_loss == Decimal("200.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2014, 9, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("500.0")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("600.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("3500.0")
    assert holding.quantity == Decimal("35.0")


def test_bed_and_breakfast_rule_with_share_split(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 20, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("150.0"),
        quantity=Decimal("5.0"),
    )

    order3 = Acquisition(
        datetime(2019, 1, 30, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("160.0"),
        quantity=Decimal("20.0"),
    )

    taxcalc = make_tax_calculator(
        [order1, order2, order3],
        [Split(datetime(2019, 1, 25, tzinfo=timezone.utc), Decimal("3.0"))],
    )
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.acquisition_date == order3.date
    assert cg.cost == Decimal("120.0")
    assert cg.quantity == Decimal("5.0")
    assert cg.gain_loss == Decimal("30.0")

    holding = taxcalc.holding(Ticker("X"))
    assert holding.cost == Decimal("140.0")
    assert holding.quantity == Decimal("35.0")


def test_get_holding_value(make_tax_calculator):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    taxcalc = make_tax_calculator([order], price=Money("15.0", GBP))
    assert taxcalc.get_holding_value(ISIN("X")) == Decimal("150.0")


def test_get_holding_value_with_currency_conversion(make_tax_calculator):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    taxcalc = make_tax_calculator(
        [order],
        price=Money("15.0", USD),
        live_rate=Decimal("0.75"),
    )
    assert taxcalc.get_holding_value(ISIN("X")) == Decimal("112.5")


def test_get_holding_value_when_price_or_fx_rate_not_available(
    make_tax_calculator,
):
    order = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        ticker="TICKER",
        total=sterling("100.0"),
        quantity=Decimal("10.0"),
    )

    taxcalc = make_tax_calculator([order], price=RequestError)
    assert taxcalc.get_holding_value(ISIN("X")) is None


#######################################################################
# Tests based on calculation examples provided online
#######################################################################


# https://assets.publishing.service.gov.uk/media/65f993439316f5001164c2d7/HS284_Example_3_2024.pdf
def test_hmrc_example_section_104_disposal(make_tax_calculator):
    order1 = Acquisition(
        datetime(2015, 4, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        total=sterling("4150.0"),
        quantity=Decimal("1000.0"),
        fees=Fees(stamp_duty=sterling("150.0")),
    )

    order2 = Acquisition(
        datetime(2018, 9, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        total=sterling("2130"),
        quantity=Decimal("500.0"),
        fees=Fees(stamp_duty=sterling("80.0")),
    )

    order3 = Disposal(
        datetime(2023, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        total=sterling("3260.0"),
        quantity=Decimal("700.0"),
        fees=Fees(stamp_duty=sterling("100.0")),
    )

    order4 = Disposal(
        datetime(2024, 2, 1, tzinfo=timezone.utc),
        isin=ISIN("LOBS"),
        total=sterling("1975"),
        quantity=Decimal("400.0"),
        fees=Fees(stamp_duty=sterling("105.0")),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2
    assert capital_gains == taxcalc.capital_gains(2024)

    cg = capital_gains[0]
    assert cg.disposal.date == date(2023, 5, 1)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("3030.67")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("329.33")
    assert str(cg) == (
        "2023-05-01 LOBS quantity: 700.0, cost: £3030.67"
        ", proceeds: £3360.0, gain: £329.33"
        ", identification: Section 104"
    )

    cg = capital_gains[1]
    assert cg.disposal.date == date(2024, 2, 1)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("1779.67")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("300.33")

    holding = taxcalc.holding(Ticker("LOBS"))
    assert holding.cost.quantize(Decimal("0.00")) == Decimal("1674.67")
    assert holding.quantity == Decimal("400.0")


# https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22251
def test_hmrc_example_crypto22251(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        total=sterling("1000.0"),
        quantity=Decimal("100.0"),
    )

    order2 = Acquisition(
        datetime(2020, 9, 18, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        total=sterling("125000.0"),
        quantity=Decimal("50.0"),
    )

    order3 = Disposal(
        datetime(2020, 12, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-A"),
        total=sterling("300000.0"),
        quantity=Decimal("50.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 12, 1)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("42000.0")
    assert cg.gain_loss == Decimal("258000.0")

    holding = taxcalc.holding(ISIN("TOKEN-A"))
    assert holding.cost == Decimal("84000.0")
    assert holding.quantity == Decimal("100.0")


# https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22252
def test_hmrc_example_crypto22252(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        total=sterling("500.0"),
        quantity=Decimal("5000.0"),
    )

    order2 = Disposal(
        datetime(2020, 6, 23, 9, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        total=sterling("800.0"),
        quantity=Decimal("1000.0"),
    )

    order3 = Acquisition(
        datetime(2020, 6, 23, 13, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        total=sterling("1000.0"),
        quantity=Decimal("1600.0"),
    )

    order4 = Disposal(
        datetime(2020, 6, 23, 19, 0, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-B"),
        total=sterling("600.0"),
        quantity=Decimal("500.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 1

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 6, 23)
    assert cg.acquisition_date == date(2020, 6, 23)
    assert cg.cost == Decimal("937.5")
    assert cg.gain_loss == Decimal("462.5")

    holding = taxcalc.holding(ISIN("TOKEN-B"))
    assert holding.cost == Decimal("562.5")
    assert holding.quantity == Decimal("5100.0")


# https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22253
def test_hmrc_example_crypto22253(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("1000.0"),
        quantity=Decimal("2000.0"),
    )

    order2 = Disposal(
        datetime(2020, 3, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("400.0"),
        quantity=Decimal("1000.0"),
    )

    order3 = Disposal(
        datetime(2020, 4, 20, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("150.0"),
        quantity=Decimal("500.0"),
    )

    order4 = Acquisition(
        datetime(2020, 4, 21, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("175.0"),
        quantity=Decimal("700.0"),
    )

    order5 = Acquisition(
        datetime(2020, 4, 28, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("100.0"),
        quantity=Decimal("500.0"),
    )

    order6 = Acquisition(
        datetime(2020, 5, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-C"),
        total=sterling("150.0"),
        quantity=Decimal("500.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5, order6])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 4

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 3, 31)
    assert cg.acquisition_date == date(2020, 4, 21)
    assert cg.cost == Decimal("175.0")
    assert cg.gain_loss == Decimal("105.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 3, 31)
    assert cg.acquisition_date == date(2020, 4, 28)
    assert cg.cost == Decimal("60.0")
    assert cg.gain_loss == Decimal("60.0")

    cg = capital_gains[2]
    assert cg.disposal.date == date(2020, 4, 20)
    assert cg.acquisition_date == date(2020, 4, 28)
    assert cg.cost == Decimal("40.0")
    assert cg.gain_loss == Decimal("20.0")

    cg = capital_gains[3]
    assert cg.disposal.date == date(2020, 4, 20)
    assert cg.acquisition_date == date(2020, 5, 1)
    assert cg.cost == Decimal("90.0")
    assert cg.gain_loss == Decimal("0.0")

    holding = taxcalc.holding(ISIN("TOKEN-C"))
    assert holding.cost == Decimal("1060.0")
    assert holding.quantity == Decimal("2200.0")


#  https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22254
def test_hmrc_example_crypto22254(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("1000.0"),
        quantity=Decimal("8000.0"),
    )

    order2 = Disposal(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("500.0"),
        quantity=Decimal("5000.0"),
    )

    order3 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("320.0"),
        quantity=Decimal("4000.0"),
    )

    order4 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("75.0"),
        quantity=Decimal("1000.0"),
    )

    order5 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("70.0"),
        quantity=Decimal("1000.0"),
    )

    order6 = Disposal(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("142.0"),
        quantity=Decimal("2000.0"),
    )

    order7 = Acquisition(
        datetime(2020, 1, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-D"),
        total=sterling("35.0"),
        quantity=Decimal("500.0"),
    )

    taxcalc = make_tax_calculator(
        [order1, order2, order3, order4, order5, order6, order7]
    )
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 1, 31)
    assert cg.acquisition_date == date(2020, 1, 31)
    assert cg.cost == Decimal("500.0")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("96.14")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 1, 31)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("62.5")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("-16.64")

    holding = taxcalc.holding(ISIN("TOKEN-D"))
    assert holding.cost == Decimal("937.5")
    assert holding.quantity == Decimal("7500.0")


# https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22255
def test_hmrc_example_crypto22255(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        total=sterling("200000.0"),
        quantity=Decimal("14000.0"),
    )

    order2 = Disposal(
        datetime(2020, 8, 30, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        total=sterling("160000.0"),
        quantity=Decimal("4000.0"),
    )

    order3 = Acquisition(
        datetime(2020, 9, 11, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-E"),
        total=sterling("17500.0"),
        quantity=Decimal("500.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 2

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 8, 30)
    assert cg.acquisition_date == date(2020, 9, 11)
    assert cg.cost == Decimal("17500.0")
    assert cg.gain_loss == Decimal("2500.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 8, 30)
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("50000.0")
    assert cg.gain_loss == Decimal("90000.0")

    holding = taxcalc.holding(ISIN("TOKEN-E"))
    assert holding.cost == Decimal("150000.0")
    assert holding.quantity == Decimal("10500.0")


# https://www.gov.uk/hmrc-internal-manuals/cryptoassets-manual/crypto22256
def test_hmrc_example_crypto22256(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("300000.0"),
        quantity=Decimal("100000.0"),
    )

    order2 = Acquisition(
        datetime(2020, 7, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("45000.0"),
        quantity=Decimal("10000.0"),
    )

    order3 = Disposal(
        datetime(2020, 7, 31, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("150000.0"),
        quantity=Decimal("30000.0"),
    )

    order4 = Disposal(
        datetime(2020, 8, 5, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("100000.0"),
        quantity=Decimal("20000.0"),
    )

    order5 = Acquisition(
        datetime(2020, 8, 6, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("225000.0"),
        quantity=Decimal("50000.0"),
    )

    order6 = Disposal(
        datetime(2020, 8, 7, tzinfo=timezone.utc),
        isin=ISIN("TOKEN-F"),
        total=sterling("150000.0"),
        quantity=Decimal("100000.0"),
    )

    taxcalc = make_tax_calculator([order1, order2, order3, order4, order5, order6])
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 4

    cg = capital_gains[0]
    assert cg.disposal.date == date(2020, 7, 31)
    assert cg.acquisition_date == date(2020, 7, 31)
    assert cg.cost == Decimal("45000.0")
    assert cg.gain_loss == Decimal("5000.0")

    cg = capital_gains[1]
    assert cg.disposal.date == date(2020, 7, 31)
    assert cg.acquisition_date == date(2020, 8, 6)
    assert cg.cost == Decimal("90000.0")
    assert cg.gain_loss == Decimal("10000.0")

    cg = capital_gains[2]
    assert cg.disposal.date == date(2020, 8, 5)
    assert cg.acquisition_date == date(2020, 8, 6)
    assert cg.cost == Decimal("90000.0")
    assert cg.gain_loss == Decimal("10000.0")

    cg = capital_gains[3]
    assert cg.disposal.date == date(2020, 8, 7)
    assert cg.acquisition_date is None
    assert cg.cost.quantize(Decimal("0.00")) == Decimal("313636.36")
    assert cg.gain_loss.quantize(Decimal("0.00")) == Decimal("-163636.36")

    holding = taxcalc.holding(ISIN("TOKEN-F"))
    assert holding.cost.quantize(Decimal("0.00")) == Decimal("31363.64")
    assert holding.quantity == Decimal("10000.0")


# https://rppaccounts.co.uk/taxation-of-shares/
def test_rawlinson_pryde_and_partners_accounts_example(make_tax_calculator):
    order1 = Acquisition(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("12025.0"),
        quantity=Decimal("4.0"),
    )

    order2 = Disposal(
        datetime(2019, 1, 18, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("9100.0"),
        quantity=Decimal("3.0"),
    )

    order3 = Disposal(
        datetime(2019, 1, 19, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("5000.0"),
        quantity=Decimal("1.0"),
    )

    order4 = Acquisition(
        datetime(2019, 1, 22, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("2000.0"),
        quantity=Decimal("1.0"),
    )

    order5 = Acquisition(
        datetime(2019, 1, 23, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("15000.0"),
        quantity=Decimal("5.0"),
    )

    order6 = Disposal(
        datetime(2019, 1, 24, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("8000.0"),
        quantity=Decimal("2.0"),
    )

    order7 = Acquisition(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("3300.0"),
        quantity=Decimal("1.0"),
    )

    order8 = Disposal(
        datetime(2019, 1, 25, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("3500.0"),
        quantity=Decimal("1.0"),
    )

    order9 = Acquisition(
        datetime(2019, 1, 26, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("17000.0"),
        quantity=Decimal("4.0"),
    )

    order10 = Disposal(
        datetime(2019, 1, 27, tzinfo=timezone.utc),
        isin=ISIN("X"),
        total=sterling("70000.0"),
        quantity=Decimal("8.0"),
    )

    taxcalc = make_tax_calculator(
        [
            order1,
            order2,
            order3,
            order4,
            order5,
            order6,
            order7,
            order8,
            order9,
            order10,
        ]
    )
    capital_gains = taxcalc.capital_gains()
    assert len(capital_gains) == 5

    cg = capital_gains[0]
    assert cg.acquisition_date == date(2019, 1, 18)
    assert cg.cost == Decimal("9018.75")
    assert cg.gain_loss == Decimal("81.25")

    cg = capital_gains[1]
    assert cg.acquisition_date == date(2019, 1, 22)
    assert cg.cost == Decimal("2000.00")
    assert cg.gain_loss == Decimal("3000.0")

    cg = capital_gains[2]
    assert cg.acquisition_date == date(2019, 1, 26)
    assert cg.cost == Decimal("8500.0")
    assert cg.gain_loss == Decimal("-500.0")

    cg = capital_gains[3]
    assert cg.acquisition_date == date(2019, 1, 25)
    assert cg.cost == Decimal("3300.0")
    assert cg.gain_loss == Decimal("200.0")

    cg = capital_gains[4]
    assert cg.acquisition_date is None
    assert cg.cost == Decimal("26506.25")
    assert cg.gain_loss == Decimal("43493.75")

    assert taxcalc.holding(Ticker("X")) is None



from datetime import date, datetime, timezone
from decimal import Decimal

import pytest

from investir.fees import Fees
from investir.findata import Split
from investir.transaction import Acquisition, Disposal, Order
from investir.typing import ISIN, Ticker
from investir.utils import sterling


def test_acquisition_order():
    count = Order.order_count

    order = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        quantity=Decimal("20.0"),
        total=sterling("101.4"),
        fees=Fees(stamp_duty=sterling("1.4")),
        tr_id="ORDER",
    )

    assert order.date == date(2022, 4, 6)
    assert order.tax_year() == 2023
    assert order.number == count + 1
    assert order.price == order.cost_before_fees / order.quantity
    assert order.cost_before_fees == order.total - order.fees.total


def test_disposal_order():
    count = Order.order_count

    order = Disposal(
        datetime(2023, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        quantity=Decimal("10.0"),
        total=sterling("48.3"),
        fees=Fees(stamp_duty=sterling("1.7")),
        tr_id="ORDER",
    )

    assert order.date == date(2023, 4, 6)
    assert order.tax_year() == 2024
    assert order.number == count + 1
    assert order.price == order.gross_proceeds / order.quantity
    assert order.gross_proceeds == order.total + order.fees.total


def test_order_merge():
    order1 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        total=sterling("101.4"),
        quantity=Decimal("20.0"),
        fees=Fees(stamp_duty=sterling("1.4")),
    )

    order2 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        total=sterling("51.5"),
        quantity=Decimal("5.0"),
        fees=Fees(stamp_duty=sterling("1.5")),
    )

    order3 = Acquisition(
        datetime(2022, 4, 6, 18, 4, 50),
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        total=sterling("10.1"),
        quantity=Decimal("4.0"),
        fees=Fees(stamp_duty=sterling("0.1")),
    )

    merged_order = Order.merge(order1, order2, order3)
    assert merged_order.timestamp == datetime(2022, 4, 6, 0, 0)
    assert merged_order.isin == ISIN("AMZN-ISIN")
    assert merged_order.ticker == Ticker("AMZN")
    assert merged_order.name == "Amazon"
    assert merged_order.total == sterling("163.0")
    assert merged_order.quantity == Decimal("29.0")
    assert merged_order.fees.total == sterling("3.0")


def test_order_split():
    date_time = datetime(2022, 4, 6, 18, 4, 50)

    order = Acquisition(
        date_time,
        isin=ISIN("AMZN-ISIN"),
        ticker=Ticker("AMZN"),
        name="Amazon",
        total=sterling("126.0"),
        quantity=Decimal("12.0"),
        fees=Fees(stamp_duty=sterling("6.0")),
    )

    matched, remainder = order.split(Decimal("8.0"))

    assert isinstance(matched, Acquisition)
    assert matched.timestamp == date_time
    assert matched.isin == ISIN("AMZN-ISIN")
    assert matched.ticker == Ticker("AMZN")
    assert matched.name == "Amazon"
    assert matched.total == sterling("84.0")
    assert matched.quantity == Decimal("8.0")
    assert matched.fees.total == sterling("4.0")

    assert isinstance(remainder, Acquisition)
    assert remainder.timestamp == date_time
    assert remainder.isin == ISIN("AMZN-ISIN")
    assert remainder.ticker == Ticker("AMZN")
    assert remainder.name == "Amazon"
    assert remainder.total == sterling("42.0")
    assert remainder.quantity == Decimal("4.0")
    assert remainder.fees.total == sterling("2.0")

    with pytest.raises(AssertionError):
        order.split(Decimal("12.1"))


def test_order_adjust_quantity():
    order1 = Acquisition(
        datetime(2018, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("AMZN-ISIN"),
        name="Amazon",
        total=sterling("10.0"),
        quantity=Decimal("1.0"),
    )

    order2 = Acquisition(
        datetime(2020, 1, 1, tzinfo=timezone.utc),
        isin=ISIN("AMZN-ISIN"),
        name="Amazon",
        total=sterling("10.0"),
        quantity=Decimal("1.0"),
    )

    splits = [
        Split(datetime(2019, 1, 1, tzinfo=timezone.utc), Decimal("10.0")),
        Split(datetime(2021, 1, 1, tzinfo=timezone.utc), Decimal("3.0")),
    ]

    order2_adjusted = order2.adjust_quantity(splits)
    ratio = Decimal("3.0")
    assert type(order2_adjusted) is type(order2)
    assert order2_adjusted.timestamp == order2.timestamp
    assert order2_adjusted.isin == order2.isin
    assert order2_adjusted.ticker == order2.ticker
    assert order2_adjusted.name == order2.name
    assert order2_adjusted.total == order2.total
    assert order2_adjusted.fees == order2.fees
    assert order2_adjusted.quantity == order2.quantity * ratio
    assert order2_adjusted.original_quantity == order2.quantity
    assert order2_adjusted.tr_id == order2.tr_id
    assert "Adjusted from order" in order2_adjusted.notes

    order1_adjusted = order1.adjust_quantity(splits)
    ratio = Decimal("10.0") * Decimal("3.0")
    assert type(order1_adjusted) is type(order1)
    assert order1_adjusted.timestamp == order1.timestamp
    assert order1_adjusted.isin == order1.isin
    assert order1_adjusted.ticker == order1.ticker
    assert order1_adjusted.name == order1.name
    assert order1_adjusted.total == order1.total
    assert order1_adjusted.fees == order1.fees
    assert order1_adjusted.quantity == order1.quantity * ratio
    assert order1_adjusted.original_quantity == order1.quantity
    assert order1_adjusted.tr_id == order1.tr_id
    assert "Adjusted from order" in order1_adjusted.notes



from datetime import date, datetime
from decimal import Decimal
from typing import Protocol

from moneyed import Currency, Money

from investir.exceptions import InvestirError
from investir.findata.types import SecurityInfo
from investir.typing import ISIN


class ProviderError(InvestirError):
    pass


class RequestError(ProviderError):
    pass


class DataNotFoundError(ProviderError):
    pass


class CacheMissError(ProviderError):
    pass


class SecurityInfoProvider(Protocol):
    def get_info(
        self, isin: ISIN, name: str = "", refresh_date: datetime | None = None
    ) -> SecurityInfo:
        pass

    def get_price(self, isin: ISIN, name: str = "") -> Money:
        pass


class LiveExchangeRateProvider(Protocol):
    def get_rate(self, base: Currency, quote: Currency) -> Decimal:
        pass


class HistoricalExchangeRateProvider(Protocol):
    def get_rate(self, base: Currency, quote: Currency, rate_date: date) -> Decimal:
        pass



from datetime import datetime, timezone
from typing import Final

from moneyed import get_currency

# Minimum timestamp for an order. Different rules apply on orders made
# before 6 April 2008. See:
# https://www.gov.uk/hmrc-internal-manuals/capital-gains-manual/cg51570
MIN_TIMESTAMP: Final = datetime(2008, 4, 6, tzinfo=timezone.utc)

CURRENCY_CODES: Final = frozenset(
    [
        "GBP",  # Pound sterling
        "USD",  # United States dollar
        "EUR",  # Euro
        "CHF",  # Swiss franc
        "DKK",  # Danish Krone
        "NOK",  # Norwegian Krone
        "PLN",  # Polish złoty
        "SEK",  # Swedish Krona
        "CZK",  # Czech Koruna
        "RON",  # Romanian Leu
        "BGN",  # Bulgarian Lev
        "HUF",  # Hungarian Forint
    ]
)

BASE_CURRENCY: Final = get_currency("GBP")



import math
from collections import defaultdict
from collections.abc import Callable, Sequence
from dataclasses import KW_ONLY, dataclass
from datetime import date
from decimal import Decimal
from enum import Enum
from functools import cache
from typing import Any, Mapping, Set

import prettytable
from moneyed import Currency, Money

from investir.const import BASE_CURRENCY
from investir.utils import boldify


class OutputFormat(str, Enum):
    TEXT = "text"
    CSV = "csv"
    JSON = "json"
    HTML = "html"


class Format(Enum):
    DATE = 1
    DECIMAL = 2
    QUANTITY = 3
    MONEY = 4


def date_format(format: str) -> Callable[[str, Any], str]:
    def _date_format(_field, val) -> str:
        if isinstance(val, date):
            return val.strftime(format)
        return val

    return _date_format


def decimal_format(precision: int) -> Callable[[str, Any], str]:
    def _decimal_format(_field, val) -> str:
        if isinstance(val, Decimal):
            return f"{val:.{precision}f}"
        elif isinstance(val, str):
            return val
        else:
            return ""

    return _decimal_format


def money_format(show_currency: bool) -> Callable[[str, Any], str]:
    def _money_format(_field, val) -> str:
        if isinstance(val, Money):
            precision = currency_precision(val.currency)
            if show_currency:
                return f"{val.amount:.{precision}f} {val.currency}"
            else:
                return f"{val.amount:.{precision}f}"
        elif isinstance(val, str):
            return val
        else:
            return ""

    return _money_format


@cache
def currency_precision(currency: Currency) -> int:
    return int(math.log10(currency.sub_unit))


@dataclass
class Field:
    name: str
    format: Format | None = None
    _: KW_ONLY
    visible: bool = True
    show_sum: bool = False


class PrettyTable(prettytable.PrettyTable):
    def __init__(
        self,
        fields: Sequence[Field],
        **kwargs,
    ) -> None:
        super().__init__([field.name for field in fields], **kwargs)

        self.hrules = prettytable.HEADER
        self.vrules = prettytable.NONE

        self.__fields = fields

    def __bool__(self) -> bool:
        return len(self.rows) > 0

    # @override
    def get_csv_string(self, **kwargs) -> str:
        import csv
        import io

        options = self._get_options(kwargs)
        csv_options = {
            key: value for key, value in kwargs.items() if key not in options
        }
        csv_buffer = io.StringIO()
        csv_writer = csv.writer(csv_buffer, **csv_options)

        if options.get("header"):
            csv_writer.writerow(self._get_expanded_field_names())

        for row in self._get_expanded_rows(options):
            csv_writer.writerow(row)

        return csv_buffer.getvalue()

    # @override
    def get_json_string(self, **kwargs) -> str:
        import json

        options = self._get_options(kwargs)
        json_options: Any = {"indent": 4, "separators": (",", ": "), "sort_keys": True}
        json_options.update(
            {key: value for key, value in kwargs.items() if key not in options}
        )
        objects: list[Any] = []

        field_names = self._get_expanded_field_names()

        if options.get("header"):
            objects.append(field_names)

        for row in self._get_expanded_rows(options):
            objects.append(dict(zip(field_names, row, strict=False)))

        return json.dumps(objects, **json_options)

    def to_string(
        self, format: OutputFormat = OutputFormat.TEXT, leading_nl: bool = True
    ) -> str:
        if format in (OutputFormat.TEXT, OutputFormat.HTML):
            self._set_fields_names(bold_text=format == OutputFormat.TEXT)

        self._set_fields_format()

        if (
            self.rows
            and any(field.show_sum for field in self.__fields)
            and format in (OutputFormat.TEXT, OutputFormat.HTML)
        ):
            self._add_total_row()

        start_nl = "\n" if leading_nl else ""
        end_nl = "\n" if format == OutputFormat.TEXT else ""

        fields = [field.name for field in self.__fields if field.visible]

        kwargs: dict[str, Any] = {"fields": fields}
        if format == OutputFormat.JSON:
            kwargs["default"] = str

        table_str = self.get_formatted_string(format, **kwargs)

        if format == OutputFormat.CSV:
            table_str = table_str.rstrip()

        return f"{start_nl}{table_str}{end_nl}"

    def _set_fields_names(self, bold_text: bool) -> None:
        for field in self.__fields:
            if field.format == Format.MONEY:
                currencies = self._get_currencies(field.name)
                if len(currencies) == 0:
                    field.name += f" ({BASE_CURRENCY})"
                elif len(currencies) == 1:
                    currency = next(iter(currencies))
                    field.name += f" ({currency.code})"

            if bold_text:
                field.name = boldify(field.name)

        self.field_names = [field.name for field in self.__fields]

    def _set_fields_format(self) -> None:
        for field in self.__fields:
            match field.format:
                case Format.DATE:
                    self.custom_format[field.name] = date_format("%d/%m/%Y")
                    self.align[field.name] = "l"
                case Format.DECIMAL:
                    self.custom_format[field.name] = decimal_format(2)
                    self.align[field.name] = "r"
                case Format.QUANTITY:
                    self.custom_format[field.name] = decimal_format(8)
                    self.align[field.name] = "r"
                case Format.MONEY:
                    self.custom_format[field.name] = money_format(
                        show_currency=self._is_multicurrency(field.name)
                    )
                    self.align[field.name] = "r"
                case _:
                    self.align[field.name] = "l"

    def _add_total_row(self) -> None:
        self.add_row(
            [
                self._sum_field(field) if field.show_sum else ""
                for field in self.__fields
            ]
        )

    def _sum_field(self, field: Field) -> str:
        idx = self._field_index(field.name)

        # TODO: Round money objects based on their subunit instead of
        #       hardcoding the number of decimals to 2.

        if field.format == Format.MONEY and self._is_multicurrency(field.name):
            totals: dict[Currency, Decimal] = defaultdict(Decimal)
            for row in self.rows:
                if row[idx]:
                    totals[row[idx].currency] += round(row[idx].amount, 2)

            return "\n".join(f"{total} {cur}" for cur, total in totals.items())

        total = sum(
            (
                round(row[idx].amount if isinstance(row[idx], Money) else row[idx], 2)
                for row in self.rows
                if row[idx] and row[idx] != "n/a"
            ),
        )

        return f"{total:.2f}"

    def _field_index(self, field_name: str) -> int:
        return next(
            idx for idx, field in enumerate(self.__fields) if field.name == field_name
        )

    def _get_currencies(self, field_name: str) -> Set[Currency]:
        idx = self._field_index(field_name)
        return set(row[idx].currency for row in self.rows if row[idx])

    def _is_multicurrency(self, field_name: str) -> bool:
        return len(self._get_currencies(field_name)) > 1

    def _get_expanded_field_names(self) -> Sequence[str]:
        field_names = []

        for field in self.__fields:
            field_names.append(field.name)
            if field.format == Format.MONEY:
                field_names.append(f"{field.name} (Currency)")

        return field_names

    def _get_expanded_rows(self, options: Mapping) -> Sequence[Sequence[Any]]:
        return [self._get_expanded_row(row) for row in self._get_rows(options)]

    def _get_expanded_row(self, row: Sequence[Any]) -> Sequence[Any]:
        out_row: list[Any] = []

        for idx, field in enumerate(row):
            if self.__fields[idx].format == Format.MONEY:
                if isinstance(field, Money):
                    out_row += [field.amount, field.currency]
                else:
                    out_row += [field, None]
            else:
                out_row.append(field)

        return out_row



import json
import logging
import os
from datetime import date, datetime
from decimal import Decimal
from pathlib import Path

import yaml
import yfinance
from moneyed import Currency, Money

from investir.config import config
from investir.findata.dataprovider import (
    CacheMissError,
    DataNotFoundError,
    RequestError,
)
from investir.findata.types import SecurityInfo, Split
from investir.typing import ISIN

logger = logging.getLogger(__name__)


def make_symbol(a: Currency, b: Currency) -> str:
    return f"{a.code}{b.code}=X"


class YahooFinanceSecurityInfoProvider:
    def __init__(self, cache_file: Path | None = None) -> None:
        self._cache_file = cache_file or config.cache_dir / "securities.yaml"
        self._cache_loaded = False
        self._infos: dict[ISIN, SecurityInfo] = {}
        self._prices: dict[ISIN, Money] = {}

    def get_info(
        self, isin: ISIN, name: str = "", refresh_date: datetime | None = None
    ) -> SecurityInfo:
        self._load_cache()

        if info := self._infos.get(isin):
            if refresh_date is None or info.last_updated >= refresh_date:
                return info

        if config.offline:
            raise CacheMissError

        logger.info("Fetching information for %s - %s", isin, name)

        try:
            yf_data = yfinance.Ticker(isin)
            name = yf_data.info["shortName"]
        except Exception as ex:
            logger.debug("Exception from yfinance: %s", repr(ex))
            raise RequestError(f"Failed to fetch information for {isin}") from None

        splits = [
            Split(pd_date.to_pydatetime(), Decimal(ratio))
            for pd_date, ratio in yf_data.splits.items()
        ]

        self._infos[isin] = SecurityInfo(name, splits)
        self._save_cache()

        return self._infos[isin]

    def get_price(self, isin: ISIN, name: str = "") -> Money:
        if cached_price := self._prices.get(isin):
            return cached_price

        if config.offline:
            raise CacheMissError

        logger.info("Fetching last price for %s - %s", isin, name)

        try:
            yf_data = yfinance.Ticker(isin)
            price = Decimal(yf_data.info["currentPrice"])
            currency = yf_data.info["currency"]
        except Exception as e:
            logger.debug("Exception from yfinance: %s", repr(e))
            raise RequestError(f"Failed to fetch last price for {isin}") from None

        if currency == "GBp":
            currency = "GBP"
            price *= Decimal("0.01")

        logger.debug(
            "Using %s %s share price for %s",
            round(price, 2),
            currency,
            isin,
        )

        self._prices[isin] = Money(price, currency)

        return self._prices[isin]

    def _save_cache(self) -> None:
        self._cache_file.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = self._cache_file.parent / (self._cache_file.name + ".tmp")

        infos = dict(sorted(self._infos.items()))
        data = {"version": 1, "securities": infos}

        with tmp_path.open("w", encoding="utf-8") as f:
            yaml.dump(data, f, sort_keys=False)

        os.replace(tmp_path, self._cache_file)

    def _load_cache(self) -> None:
        if not self._cache_loaded and self._cache_file.exists():
            logger.info("Loading securities cache from %s", self._cache_file)
            with self._cache_file.open(encoding="utf-8") as f:
                if data := yaml.load(f, Loader=yaml.FullLoader):
                    self._infos = data["securities"]

        self._cache_loaded = True


class YahooFinanceLiveExchangeRateProvider:
    def __init__(self) -> None:
        self._rates: dict[tuple[Currency, Currency], Decimal] = {}

    def get_rate(self, base: Currency, quote: Currency) -> Decimal:
        if rate := self._rates.get((base, quote)):
            return rate

        if config.offline:
            raise CacheMissError

        try:
            yf_data = yfinance.Ticker(make_symbol(base, quote))
            rate = Decimal(yf_data.info["bid"])
        except Exception as e:
            logger.debug("Exception from yfinance: %s", repr(e))
            raise RequestError(
                f"Failed to fetch exchange rate for "
                f"{base.name} ({base.code}) to {quote.name} ({quote.code})"
            ) from None

        inverse_rate = Decimal("1.0") / rate
        logger.debug(
            "Using %s exchange rate for %s (%s) to %s (%s) (inverse rate: %s)",
            round(rate, 5),
            base.name,
            base.code,
            quote.name,
            quote.code,
            round(inverse_rate, 5),
        )

        self._rates[(base, quote)] = rate
        self._rates[(quote, base)] = inverse_rate

        return rate


class YahooFinanceHistoricalExchangeRateProvider:
    def __init__(self, cache_file: Path | None = None) -> None:
        self._cache: dict[str, dict[str, str]] = {}
        self._cache_file = cache_file or config.cache_dir / "yahoo-finance-rates.json"
        self._cache_loaded = False

    def get_rate(self, base: Currency, quote: Currency, rate_date: date) -> Decimal:
        self._load_cache()

        if rate := self._find_rate(base, quote, rate_date):
            return rate

        if config.offline:
            raise CacheMissError

        try:
            ticker = yfinance.Ticker(make_symbol(base, quote))
            rates = ticker.history(start=str(rate_date))
        except Exception as e:
            logger.debug("Exception from yfinance: %s", repr(e))
            raise RequestError(
                f"Failed to fetch exchange rate for "
                f"{base.name} ({base.code}) to "
                f"{quote.name} ({quote.code})"
            ) from None

        for timestamp, row in rates.iterrows():
            currency_pair = f"{base.code}-{quote.code}"
            date_key = str(timestamp.to_pydatetime().date())
            self._cache.setdefault(currency_pair, {})[date_key] = str(row["Close"])

        self._save_cache()

        if (rate := self._find_rate(base, quote, rate_date)) is None:
            raise DataNotFoundError(
                f"Exchange rate not found: {base.code}-{quote.code}"
            )

        return rate

    def _find_rate(
        self, base: Currency, quote: Currency, rate_date: date
    ) -> Decimal | None:
        if rates := self._cache.get(f"{base.code}-{quote.code}"):
            if rate := rates.get(str(rate_date)):
                return Decimal(rate)
        elif rates := self._cache.get(f"{quote.code}-{base.code}"):
            if rate := rates.get(str(rate_date)):
                return Decimal("1.0") / Decimal(rate)

        return None

    def _save_cache(self) -> None:
        self._cache_file.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = self._cache_file.parent / (self._cache_file.name + ".tmp")

        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(self._cache, f, indent=4)

        os.replace(tmp_path, self._cache_file)

    def _load_cache(self) -> None:
        if not self._cache_loaded and self._cache_file.exists():
            logger.info("Loading historical exchange rates from %s", self._cache_file)
            with self._cache_file.open(encoding="utf-8") as f:
                self._cache.update(json.load(f))

        self._cache_loaded = True



import logging
from datetime import date, datetime
from decimal import Decimal

from moneyed import Currency, Money

from investir.findata.dataprovider import (
    CacheMissError,
    HistoricalExchangeRateProvider,
    LiveExchangeRateProvider,
    ProviderError,
    SecurityInfoProvider,
)
from investir.findata.types import SecurityInfo
from investir.typing import ISIN

logger = logging.getLogger(__name__)


class FinancialData:
    def __init__(
        self,
        security_info_provider: SecurityInfoProvider | None,
        live_rates_provider: LiveExchangeRateProvider | None,
        historical_rates_provider: HistoricalExchangeRateProvider | None,
    ) -> None:
        self._security_info_provider = security_info_provider
        self._live_rates_provider = live_rates_provider
        self._historical_rates_provider = historical_rates_provider

    def get_security_info(
        self, isin: ISIN, name: str = "", refresh_date: datetime | None = None
    ) -> SecurityInfo:
        if self._security_info_provider is not None:
            try:
                return self._security_info_provider.get_info(isin, name, refresh_date)
            except CacheMissError:
                pass
            except ProviderError as ex:
                logger.warning(str(ex))

        return SecurityInfo(name, [])

    def get_security_price(self, isin, name: str = "") -> Money | None:
        if self._security_info_provider is not None:
            try:
                return self._security_info_provider.get_price(isin, name)
            except CacheMissError:
                pass
            except ProviderError as ex:
                logger.warning(str(ex))

        return None

    def get_exchange_rate(
        self, base: Currency, quote: Currency, rate_date: date | None = None
    ) -> Decimal | None:
        try:
            if rate_date is None and self._live_rates_provider is not None:
                return self._live_rates_provider.get_rate(base, quote)
            if rate_date is not None and self._historical_rates_provider is not None:
                return self._historical_rates_provider.get_rate(base, quote, rate_date)
        except CacheMissError:
            pass
        except ProviderError as ex:
            logger.warning(str(ex))

        return None

    def convert_money(
        self, money: Money, currency, rate_date: date | None = None
    ) -> Money | None:
        if money.currency == currency:
            return money

        if rate := self.get_exchange_rate(money.currency, currency, rate_date):
            return Money(money.amount * rate, currency)

        return None



import json
import logging
import os
from codecs import iterdecode
from csv import DictReader
from datetime import date
from decimal import Decimal
from pathlib import Path
from urllib import request
from urllib.error import URLError

from moneyed import GBP, Currency

from investir.config import config
from investir.const import CURRENCY_CODES
from investir.findata.dataprovider import (
    CacheMissError,
    DataNotFoundError,
    RequestError,
)

logger = logging.getLogger(__name__)


TRADE_TARIFF_URL = (
    "https://www.trade-tariff.service.gov.uk/api/v2/exchange_rates/files/"
)


def date_to_key(d: date) -> str:
    return d.strftime("%Y-%m")


def date_to_filename(d: date) -> str:
    return f"monthly_csv_{d.strftime('%Y-%m')}.csv"


class HmrcMonthlyExhangeRateProvider:
    def __init__(self, cache_file: Path | None = None) -> None:
        self._cache: dict[str, dict[str, str]] = {}
        self._cache_file = cache_file or config.cache_dir / "hmrc-monthly-rates.json"
        self._cache_loaded = False

    def get_rate(self, base: Currency, quote: Currency, rate_date: date) -> Decimal:
        self._load_cache()

        if rate := self._find_rate(base, quote, rate_date):
            return rate

        if config.offline:
            raise CacheMissError

        filename = date_to_filename(rate_date)
        url = f"{TRADE_TARIFF_URL}{filename}"
        key = date_to_key(rate_date)

        try:
            with request.urlopen(url) as response:
                charset = response.info().get_content_charset()
                for row in DictReader(iterdecode(response, charset)):
                    currency_code = row["Currency Code"]
                    currency_rate = row["Currency Units per £1"]
                    if currency_code in CURRENCY_CODES:
                        self._cache.setdefault(key, {})[currency_code] = currency_rate
        except URLError as e:
            logger.debug("Exception from urllib: %s", repr(e))
            raise RequestError(
                f"Failed to fetch exchange rates file ({filename}): {e}"
            ) from None

        self._save_cache()

        if (rate := self._find_rate(base, quote, rate_date)) is None:
            raise DataNotFoundError(
                f"Exchange rate not found: {base.code}-{quote.code}"
            )

        return rate

    def _find_rate(
        self, base: Currency, quote: Currency, rate_date: date
    ) -> Decimal | None:
        key = date_to_key(rate_date)
        rates = self._cache.get(key, {})

        if base == GBP:
            if rate := rates.get(quote.code):
                return Decimal(rate)
        elif quote == GBP:
            if rate := rates.get(base.code):
                return Decimal("1.0") / Decimal(rate)
        else:
            raise ValueError("Either 'base' or 'quote' must be 'GBP'")

        return None

    def _save_cache(self) -> None:
        self._cache_file.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = self._cache_file.parent / (self._cache_file.name + ".tmp")

        with tmp_path.open("w", encoding="utf-8") as f:
            json.dump(self._cache, f, indent=4, sort_keys=True)

        os.replace(tmp_path, self._cache_file)

    def _load_cache(self) -> None:
        if not self._cache_loaded and self._cache_file.exists():
            logger.info("Loading historical exchange rates from %s", self._cache_file)
            with self._cache_file.open(encoding="utf-8") as f:
                self._cache = json.load(f)

        self._cache_loaded = True



import logging
from collections.abc import Mapping
from csv import DictReader
from datetime import datetime
from decimal import ROUND_DOWN, Decimal
from pathlib import Path
from typing import Final

from dateutil.parser import parse as parse_timestamp
from moneyed import GBP, Money

from investir.const import MIN_TIMESTAMP
from investir.exceptions import (
    CalculatedAmountError,
    FeesError,
    FieldUnknownError,
    OrderDateError,
    TransactionTypeError,
)
from investir.fees import Fees
from investir.parser.factory import ParserFactory
from investir.parser.types import ParsingResult
from investir.transaction import (
    Acquisition,
    Disposal,
    Dividend,
    Interest,
    Order,
    Transfer,
)
from investir.typing import ISIN, Ticker
from investir.utils import dict2str, raise_or_warn, read_decimal, read_sterling

logger = logging.getLogger(__name__)


@ParserFactory.register("Freetrade")
class FreetradeParser:
    FIELDS: Final = (
        "Title",
        "Type",
        "Timestamp",
        "Account Currency",
        "Total Amount",
        "Buy / Sell",
        "Ticker",
        "ISIN",
        "Price per Share in Account Currency",
        "Stamp Duty",
        "Quantity",
        "Venue",
        "Order ID",
        "Order Type",
        "Instrument Currency",
        "Total Shares Amount",
        "Price per Share",
        "FX Rate",
        "Base FX Rate",
        "FX Fee (BPS)",
        "FX Fee Amount",
        "Dividend Ex Date",
        "Dividend Pay Date",
        "Dividend Eligible Quantity",
        "Dividend Amount Per Share",
        "Dividend Gross Distribution Amount",
        "Dividend Net Distribution Amount",
        "Dividend Withheld Tax Percentage",
        "Dividend Withheld Tax Amount",
        "Stock Split Ex Date",
        "Stock Split Pay Date",
        "Stock Split New ISIN",
        "Stock Split Rate of Share Outturn From",
        "Stock Split Rate of Share Outturn To",
        "Stock Split Maintain Holding of Initial ISIN",
        "Stock Split New Share Quantity",
        "Stock Split Rate of Cash Outturn Amount",
        "Stock Split Rate of Cash Outturn Currency",
        "Stock Split Cash Outturn Received Amount",
        "Stock Split Has Fractional Payout",
        "Stock Split Rate of Fractional Payout Amount",
        "Stock Split Rate of Fractional Payout Currency",
        "Stock Split Fractional Payout Cash Received Amount",
        "Stock Split Fractional Payout Cash Received Currency",
    )

    REQUIRED: Final = ("Type", "Timestamp", "Total Amount", "Account Currency")

    def __init__(self, csv_file: Path) -> None:
        self._csv_file = csv_file
        self._orders: list[Order] = []
        self._dividends: list[Dividend] = []
        self._transfers: list[Transfer] = []
        self._interest: list[Interest] = []

    def can_parse(self) -> bool:
        with self._csv_file.open(encoding="utf-8") as file:
            reader = DictReader(file)
            fieldnames = reader.fieldnames or []

        return all(f in fieldnames for f in self.REQUIRED)

    def parse(self) -> ParsingResult:
        parse_fn = {
            "ORDER": self._parse_order,
            "DIVIDEND": self._parse_dividend,
            "TOP_UP": self._parse_transfer,
            "WITHDRAWAL": self._parse_transfer,
            "INTEREST_FROM_CASH": self._parse_interest,
            "MONTHLY_STATEMENT": None,
            "TAX_CERTIFICATE": None,
        }

        with self._csv_file.open(encoding="utf-8") as file:
            reader = DictReader(file)

            unknown_fields = [
                f for f in reader.fieldnames or [] if f not in self.FIELDS
            ]
            if unknown_fields:
                raise_or_warn(FieldUnknownError(unknown_fields))

            # Freetrade transactions are ordered from most recent to
            # oldest but we want the order ID to increase from the
            # oldest to the most recent.
            rows = reversed(list(reader))

            for row in rows:
                tr_type = row["Type"]

                if tr_type not in parse_fn:
                    raise_or_warn(TransactionTypeError(self._csv_file, row, tr_type))
                    continue

                if fn := parse_fn.get(tr_type):
                    timestamp = parse_timestamp(row["Timestamp"])
                    total = Money(Decimal(row["Total Amount"]), row["Account Currency"])

                    fn(row, tr_type, timestamp, total)

        return ParsingResult(
            self._orders, self._dividends, self._transfers, self._interest
        )

    def _parse_order(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        total: Money,
    ) -> None:
        title = row["Title"]
        action = row["Buy / Sell"]
        ticker = row["Ticker"]
        isin = row["ISIN"]
        price = Decimal(row["Price per Share in Account Currency"])
        quantity = Decimal(row["Quantity"])
        order_id = row["Order ID"]
        stamp_duty = read_sterling(row["Stamp Duty"])
        fx_fee = read_sterling(row["FX Fee Amount"])

        if action not in ("BUY", "SELL"):
            raise TransactionTypeError(self._csv_file, row, action)

        if timestamp < MIN_TIMESTAMP:
            raise OrderDateError(self._csv_file, row)

        if stamp_duty and fx_fee:
            raise FeesError(self._csv_file, row, "Stamp Duty", "FX Fee Amount")

        fees = Fees(
            stamp_duty=stamp_duty, forex=fx_fee, default_currency=total.currency
        )

        order_class: type[Order] = Acquisition
        fees_total = fees.total

        if action == "SELL":
            order_class = Disposal
            fees_total *= -1

        calculated_total = (Money(price * quantity, GBP) + fees_total).round(2)
        if calculated_total != total:
            raise_or_warn(
                CalculatedAmountError(
                    self._csv_file, row, total.amount, calculated_total.amount
                )
            )

        self._orders.append(
            order_class(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=title,
                total=total,
                quantity=quantity,
                fees=fees,
                tr_id=order_id,
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._orders[-1])

    def _parse_dividend(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        total: Money,
    ):
        title = row["Title"]
        ticker = row["Ticker"]
        isin = row["ISIN"]
        base_fx_rate = read_decimal(row["Base FX Rate"], Decimal("1.0"))
        eligible_quantity = Decimal(row["Dividend Eligible Quantity"])
        amount_per_share = Decimal(row["Dividend Amount Per Share"])
        withheld_tax_percentage = Decimal(row["Dividend Withheld Tax Percentage"])
        withheld_tax_amount = Decimal(row["Dividend Withheld Tax Amount"])

        calculated_total = (
            amount_per_share
            * eligible_quantity
            * ((Decimal("100") - withheld_tax_percentage) / 100)
            * base_fx_rate
        )

        calculated_total = calculated_total.quantize(
            Decimal("1.00"), rounding=ROUND_DOWN
        )

        # Freetrade does not seem to use a consistent method for rounding dividends.
        # Thus, allow the calculated amount to differ by one pence.
        # https://community.freetrade.io/t/dividend-amount-off-by-one-penny/71806/7
        if abs(total.amount - calculated_total) > Decimal("0.01"):
            raise_or_warn(
                CalculatedAmountError(
                    self._csv_file, row, total.amount, calculated_total
                )
            )

        self._dividends.append(
            Dividend(
                timestamp,
                isin=ISIN(isin),
                ticker=Ticker(ticker),
                name=title,
                total=total,
                withheld=Money(withheld_tax_amount * base_fx_rate, total.currency),
            )
        )

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._dividends[-1])

    def _parse_transfer(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        total: Money,
    ):
        if tr_type == "WITHDRAWAL":
            total = -abs(total)

        self._transfers.append(Transfer(timestamp, total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._transfers[-1])

    def _parse_interest(
        self,
        row: Mapping[str, str],
        tr_type: str,
        timestamp: datetime,
        total: Money,
    ):
        self._interest.append(Interest(timestamp, total))

        logger.debug("Parsed row %s as %s\n", dict2str(row), self._interest[-1])



import logging
from csv import DictReader
from datetime import date
from decimal import Decimal
from pathlib import Path
from typing import Final

from moneyed import Currency

from investir.const import BASE_CURRENCY
from investir.findata.dataprovider import DataNotFoundError, ProviderError

logger = logging.getLogger(__name__)


FIELDS: Final = ["Date", "Currency", "Rate"]


class LocalHistoricalExchangeRateProvider:
    def __init__(self, rates_file: Path) -> None:
        self._rates: dict[date, dict[Currency, Decimal]] = {}

        self._load_rates(rates_file)

    def get_rate(self, base: Currency, quote: Currency, rate_date: date) -> Decimal:
        if rate := self._find_rate(base, quote, rate_date):
            return rate

        raise DataNotFoundError(f"Exchange rate not found: {base.code}-{quote.code}")

    def _find_rate(
        self, base: Currency, quote: Currency, rate_date: date
    ) -> Decimal | None:
        rates = self._rates.get(rate_date, {})

        if base == BASE_CURRENCY:
            if rate := rates.get(quote):
                return Decimal(rate)
        elif quote == BASE_CURRENCY:
            if rate := rates.get(base):
                return Decimal("1.0") / Decimal(rate)
        else:
            raise ValueError(f"Either 'base' or 'quote' must be '{BASE_CURRENCY}'")

        return None

    def _load_rates(self, rates_file: Path) -> None:
        logger.info("Loading historical exchange rates from %s", rates_file)
        with rates_file.open(encoding="utf-8") as file:
            reader = DictReader(file)

            if reader.fieldnames != FIELDS:
                raise ProviderError("Exchange rates file is invalid")

            for row in reader:
                rate_date = date.fromisoformat(row["Date"])
                currency_code = Currency(row["Currency"])
                currency_rate = Decimal(row["Rate"])
                self._rates.setdefault(rate_date, {})[currency_code] = currency_rate



from decimal import Decimal

from moneyed import USD

from investir.const import BASE_CURRENCY
from investir.fees import Fees
from investir.utils import sterling

FEES = Fees(
    stamp_duty=sterling("4.0"),
    forex=sterling("3.0"),
    finra=sterling("0.4"),
    sec=sterling("0.1"),
)


def test_fees_none_set():
    assert Fees().total == BASE_CURRENCY.zero
    assert Fees(default_currency=USD).total == USD.zero


def test_fees_all_set():
    assert FEES.total == sterling("7.5")


def test_fees_all_set_add_sub():
    other_fees = Fees(
        stamp_duty=sterling("0.01"),
        forex=sterling("0.02"),
        finra=sterling("0.03"),
        sec=sterling("0.04"),
    )

    fees = FEES + other_fees
    assert fees.stamp_duty == sterling("4.01")
    assert fees.forex == sterling("3.02")
    assert fees.finra == sterling("0.43")
    assert fees.sec == sterling("0.14")

    assert fees - other_fees == FEES


def test_fees_none_set_add_sub():
    fees = Fees() - Fees()
    assert fees.stamp_duty is None
    assert fees.forex is None
    assert fees.finra is None
    assert fees.sec is None

    fees = Fees() + Fees()
    assert fees.stamp_duty is None
    assert fees.forex is None
    assert fees.finra is None
    assert fees.sec is None


def test_fees_some_set_add_sub():
    fees = FEES + Fees()
    assert fees.stamp_duty == sterling("4.0")
    assert fees.forex == sterling("3.0")
    assert fees.finra == sterling("0.4")
    assert fees.sec == sterling("0.1")

    fees = Fees() + FEES
    assert fees.stamp_duty == sterling("4.0")
    assert fees.forex == sterling("3.0")
    assert fees.finra == sterling("0.4")
    assert fees.sec == sterling("0.1")

    fees = FEES - Fees()
    assert fees.stamp_duty == sterling("4.0")
    assert fees.forex == sterling("3.0")
    assert fees.finra == sterling("0.4")
    assert fees.sec == sterling("0.1")

    fees = Fees() - FEES
    assert fees.stamp_duty == sterling("-4.0")
    assert fees.forex == sterling("-3.0")
    assert fees.finra == sterling("-0.4")
    assert fees.sec == sterling("-0.1")


def test_fees_all_set_mul_div():
    fees = FEES / Decimal("2.0")
    assert fees.stamp_duty == sterling("2.0")
    assert fees.forex == sterling("1.5")
    assert fees.finra == sterling("0.2")
    assert fees.sec == sterling("0.05")

    assert fees * Decimal("2.0") == FEES


def test_fees_none_set_mul_div():
    fees = Fees() / Decimal("2.0")
    assert fees.stamp_duty is None
    assert fees.forex is None
    assert fees.finra is None
    assert fees.sec is None

    fees = Fees() * Decimal("2.0")
    assert fees.stamp_duty is None
    assert fees.forex is None
    assert fees.finra is None
    assert fees.sec is None



from investir.findata.dataprovider import (
    CacheMissError,
    DataNotFoundError,
    HistoricalExchangeRateProvider,
    LiveExchangeRateProvider,
    ProviderError,
    RequestError,
    SecurityInfoProvider,
)
from investir.findata.financialdata import FinancialData
from investir.findata.hmrcprovider import HmrcMonthlyExhangeRateProvider
from investir.findata.localprovider import LocalHistoricalExchangeRateProvider
from investir.findata.types import SecurityInfo, Split
from investir.findata.yahoofinanceprovider import (
    YahooFinanceHistoricalExchangeRateProvider,
    YahooFinanceLiveExchangeRateProvider,
    YahooFinanceSecurityInfoProvider,
)

__all__ = [
    "CacheMissError",
    "DataNotFoundError",
    "FinancialData",
    "HistoricalExchangeRateProvider",
    "HmrcMonthlyExhangeRateProvider",
    "LiveExchangeRateProvider",
    "LocalHistoricalExchangeRateProvider",
    "ProviderError",
    "RequestError",
    "SecurityInfo",
    "SecurityInfoProvider",
    "Split",
    "YahooFinanceHistoricalExchangeRateProvider",
    "YahooFinanceLiveExchangeRateProvider",
    "YahooFinanceSecurityInfoProvider",
]
